"""Implements the :stac-ext:`Point Cloud Extension <pointcloud>`."""

from __future__ import annotations

from collections.abc import Iterable
from typing import (
    Any,
    Generic,
    Literal,
    TypeVar,
    cast,
)

import pystac
from pystac.extensions.base import (
    ExtensionManagementMixin,
    PropertiesExtension,
    SummariesExtension,
)
from pystac.extensions.hooks import ExtensionHooks
from pystac.summaries import RangeSummary
from pystac.utils import StringEnum, get_required, map_opt

#: Generalized version of :class:`~pystac.Item`, :class:`~pystac.Asset`,
#: or :class:`~pystac.ItemAssetDefinition`
T = TypeVar("T", pystac.Item, pystac.Asset, pystac.ItemAssetDefinition)

SCHEMA_URI: str = "https://stac-extensions.github.io/pointcloud/v1.0.0/schema.json"
PREFIX: str = "pc:"

COUNT_PROP = PREFIX + "count"
TYPE_PROP = PREFIX + "type"
ENCODING_PROP = PREFIX + "encoding"
SCHEMAS_PROP = PREFIX + "schemas"
DENSITY_PROP = PREFIX + "density"
STATISTICS_PROP = PREFIX + "statistics"


class PhenomenologyType(StringEnum):
    """Valid values for the ``pc:type`` field in the :stac-ext:`Pointcloud Item
    Properties <pointcloud#item-properties>`."""

    LIDAR = "lidar"
    EOPC = "eopc"
    RADAR = "radar"
    SONAR = "sonar"
    OTHER = "other"


class SchemaType(StringEnum):
    """Valid values for the ``type`` field in a :stac-ext:`Schema Object
    <pointcloud#schema-object>`."""

    FLOATING = "floating"
    UNSIGNED = "unsigned"
    SIGNED = "signed"


class Schema:
    """Defines a schema for dimension of a pointcloud (e.g., name, size, type)

    Use :meth:`Schema.create` to create a new instance of ``Schema`` from
    properties.
    """

    properties: dict[str, Any]

    def __init__(self, properties: dict[str, Any]) -> None:
        self.properties = properties

    def apply(self, name: str, size: int, type: SchemaType) -> None:
        """Sets the properties for this Schema.

        Args:
           name : The name of dimension.
           size : The size of the dimension in bytes. Whole bytes are supported.
           type : Dimension type. Valid values are ``floating``, ``unsigned``, and
           ``signed``
        """
        self.properties["name"] = name
        self.properties["size"] = size
        self.properties["type"] = type

    @classmethod
    def create(cls, name: str, size: int, type: SchemaType) -> Schema:
        """Creates a new Schema.

        Args:
           name : The name of dimension.
           size : The size of the dimension in bytes. Whole bytes are supported.
           type : Dimension type. Valid values are ``floating``, ``unsigned``, and
           ``signed``
        """
        c = cls({})
        c.apply(name=name, size=size, type=type)
        return c

    @property
    def size(self) -> int:
        """Gets or sets the size value."""
        return cast(int, get_required(self.properties.get("size"), self, "size"))

    @size.setter
    def size(self, v: int) -> None:
        if not isinstance(v, int):
            raise pystac.STACError(f"size must be an int! Invalid input: {v}")

        self.properties["size"] = v

    @property
    def name(self) -> str:
        """Gets or sets the name property for this Schema."""
        return cast(str, get_required(self.properties.get("name"), self, "name"))

    @name.setter
    def name(self, v: str) -> None:
        self.properties["name"] = v

    @property
    def type(self) -> SchemaType:
        """Gets or sets the type property. Valid values are ``floating``, ``unsigned``,
        and ``signed``."""
        return cast(SchemaType, get_required(self.properties.get("type"), self, "type"))

    @type.setter
    def type(self, v: SchemaType) -> None:
        self.properties["type"] = v

    def __repr__(self) -> str:
        return "<Schema name={} size={} type={}>".format(
            self.properties.get("name"),
            self.properties.get("size"),
            self.properties.get("type"),
        )

    def to_dict(self) -> dict[str, Any]:
        """Returns this schema as a dictionary."""
        return self.properties


class Statistic:
    """Defines a single statistic for Pointcloud channel or dimension

    Use :meth:`Statistic.create` to create a new instance of
    ``Statistic`` from property values."""

    properties: dict[str, Any]

    def __init__(self, properties: dict[str, Any]) -> None:
        self.properties = properties

    def apply(
        self,
        name: str,
        position: int | None = None,
        average: float | None = None,
        count: int | None = None,
        maximum: float | None = None,
        minimum: float | None = None,
        stddev: float | None = None,
        variance: float | None = None,
    ) -> None:
        """Sets the properties for this Statistic.

        Args:
            name : REQUIRED. The name of the channel.
            position : Optional position of the channel in the schema.
            average : Optional average of the channel.
            count : Optional number of elements in the channel.
            maximum : Optional maximum value of the channel.
            minimum : Optional minimum value of the channel.
            stddev : Optional standard deviation of the channel.
            variance : Optional variance of the channel.
        """
        self.properties["name"] = name
        self.properties["position"] = position
        self.properties["average"] = average
        self.properties["count"] = count
        self.properties["maximum"] = maximum
        self.properties["minimum"] = minimum
        self.properties["stddev"] = stddev
        self.properties["variance"] = variance

    @classmethod
    def create(
        cls,
        name: str,
        position: int | None = None,
        average: float | None = None,
        count: int | None = None,
        maximum: float | None = None,
        minimum: float | None = None,
        stddev: float | None = None,
        variance: float | None = None,
    ) -> Statistic:
        """Creates a new Statistic class.

        Args:
            name : REQUIRED. The name of the channel.
            position : Optional position of the channel in the schema.
            average : Optional average of the channel.
            count : Optional number of elements in the channel.
            maximum : Optional maximum value of the channel.
            minimum : Optional minimum value of the channel.
            stddev : Optional standard deviation of the channel.
            variance : Optional variance of the channel.
        """
        c = cls({})
        c.apply(
            name=name,
            position=position,
            average=average,
            count=count,
            maximum=maximum,
            minimum=minimum,
            stddev=stddev,
            variance=variance,
        )
        return c

    @property
    def name(self) -> str:
        """Gets or sets the name property."""
        return cast(str, get_required(self.properties.get("name"), self, "name"))

    @name.setter
    def name(self, v: str) -> None:
        if v is not None:
            self.properties["name"] = v
        else:
            self.properties.pop("name", None)

    @property
    def position(self) -> int | None:
        """Gets or sets the position property."""
        return self.properties.get("position")

    @position.setter
    def position(self, v: int | None) -> None:
        if v is not None:
            self.properties["position"] = v
        else:
            self.properties.pop("position", None)

    @property
    def average(self) -> float | None:
        """Gets or sets the average property."""
        return self.properties.get("average")

    @average.setter
    def average(self, v: float | None) -> None:
        if v is not None:
            self.properties["average"] = v
        else:
            self.properties.pop("average", None)

    @property
    def count(self) -> int | None:
        """Gets or sets the count property."""
        return self.properties.get("count")

    @count.setter
    def count(self, v: int | None) -> None:
        if v is not None:
            self.properties["count"] = v
        else:
            self.properties.pop("count", None)

    @property
    def maximum(self) -> float | None:
        """Gets or sets the maximum property."""
        return self.properties.get("maximum")

    @maximum.setter
    def maximum(self, v: float | None) -> None:
        if v is not None:
            self.properties["maximum"] = v
        else:
            self.properties.pop("maximum", None)

    @property
    def minimum(self) -> float | None:
        """Gets or sets the minimum property."""
        return self.properties.get("minimum")

    @minimum.setter
    def minimum(self, v: float | None) -> None:
        if v is not None:
            self.properties["minimum"] = v
        else:
            self.properties.pop("minimum", None)

    @property
    def stddev(self) -> float | None:
        """Gets or sets the stddev property."""
        return self.properties.get("stddev")

    @stddev.setter
    def stddev(self, v: float | None) -> None:
        if v is not None:
            self.properties["stddev"] = v
        else:
            self.properties.pop("stddev", None)

    @property
    def variance(self) -> float | None:
        """Gets or sets the variance property."""
        return self.properties.get("variance")

    @variance.setter
    def variance(self, v: float | None) -> None:
        if v is not None:
            self.properties["variance"] = v
        else:
            self.properties.pop("variance", None)

    def __repr__(self) -> str:
        return f"<Statistic statistics={str(self.properties)}>"

    def to_dict(self) -> dict[str, Any]:
        """Returns this statistic as a dictionary."""
        return self.properties

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, Statistic):
            return NotImplemented
        return self.to_dict() == o.to_dict()


class PointcloudExtension(
    Generic[T],
    PropertiesExtension,
    ExtensionManagementMixin[pystac.Item | pystac.Collection],
):
    """An abstract class that can be used to extend the properties of an
    :class:`~pystac.Item` or :class:`~pystac.Asset` with properties from the
    :stac-ext:`Point Cloud Extension <pointcloud>`. This class is generic over the type
    of STAC Object to be extended (e.g. :class:`~pystac.Item`,
    :class:`~pystac.Asset`).

    To create a concrete instance of :class:`PointcloudExtension`, use the
    :meth:`PointcloudExtension.ext` method. For example:

    .. code-block:: python

       >>> item: pystac.Item = ...
       >>> pc_ext = PointcloudExtension.ext(item)
    """

    name: Literal["pc"] = "pc"

    def apply(
        self,
        count: int,
        type: PhenomenologyType | str,
        encoding: str,
        schemas: list[Schema],
        density: float | None = None,
        statistics: list[Statistic] | None = None,
    ) -> None:
        """Applies Pointcloud extension properties to the extended Item.

        Args:
            count : REQUIRED. The number of points in the cloud.
            type : REQUIRED. Phenomenology type for the point cloud. Possible valid
                values might include lidar, eopc, radar, sonar, or otherThe type of file
                or data format of the cloud.
            encoding : REQUIRED. Content encoding or format of the data.
            schemas : REQUIRED. A sequential array of items
                that define the dimensions and their types.
            density : Number of points per square unit area.
            statistics : A sequential array of items mapping to
                pc:schemas defines per-channel statistics.
        """
        self.count = count
        self.type = type
        self.encoding = encoding
        self.schemas = schemas
        self.density = density
        self.statistics = statistics

    @property
    def count(self) -> int:
        """Gets or sets the number of points in the Item."""
        return get_required(self._get_property(COUNT_PROP, int), self, COUNT_PROP)

    @count.setter
    def count(self, v: int) -> None:
        self._set_property(COUNT_PROP, v, pop_if_none=False)

    @property
    def type(self) -> PhenomenologyType | str:
        """Gets or sets the phenomenology type for the point cloud."""
        return get_required(self._get_property(TYPE_PROP, str), self, TYPE_PROP)

    @type.setter
    def type(self, v: PhenomenologyType | str) -> None:
        self._set_property(TYPE_PROP, v, pop_if_none=False)

    @property
    def encoding(self) -> str:
        """Gets or sets the content encoding or format of the data."""
        return get_required(self._get_property(ENCODING_PROP, str), self, ENCODING_PROP)

    @encoding.setter
    def encoding(self, v: str) -> None:
        self._set_property(ENCODING_PROP, v, pop_if_none=False)

    @property
    def schemas(self) -> list[Schema]:
        """Gets or sets the list of :class:`Schema` instances defining
        dimensions and types for the data.
        """
        result = get_required(
            self._get_property(SCHEMAS_PROP, list[dict[str, Any]]), self, SCHEMAS_PROP
        )
        return [Schema(s) for s in result]

    @schemas.setter
    def schemas(self, v: list[Schema]) -> None:
        self._set_property(SCHEMAS_PROP, [x.to_dict() for x in v], pop_if_none=False)

    @property
    def density(self) -> float | None:
        """Gets or sets the number of points per square unit area."""
        return self._get_property(DENSITY_PROP, float)

    @density.setter
    def density(self, v: float | None) -> None:
        self._set_property(DENSITY_PROP, v)

    @property
    def statistics(self) -> list[Statistic] | None:
        """Gets or sets the list of :class:`Statistic` instances describing
        the pre-channel statistics. Elements in this list map to elements in the
        :attr:`PointcloudExtension.schemas` list."""
        result = self._get_property(STATISTICS_PROP, list[dict[str, Any]])
        return map_opt(lambda stats: [Statistic(s) for s in stats], result)

    @statistics.setter
    def statistics(self, v: list[Statistic] | None) -> None:
        set_value = map_opt(lambda stats: [s.to_dict() for s in stats], v)
        self._set_property(STATISTICS_PROP, set_value)

    @classmethod
    def get_schema_uri(cls) -> str:
        return SCHEMA_URI

    @classmethod
    def ext(cls, obj: T, add_if_missing: bool = False) -> PointcloudExtension[T]:
        """Extends the given STAC Object with properties from the :stac-ext:`Point Cloud
        Extension <pointcloud>`.

        This extension can be applied to instances of :class:`~pystac.Item` or
        :class:`~pystac.Asset`.

        Raises:

            pystac.ExtensionTypeError : If an invalid object type is passed.
        """
        if isinstance(obj, pystac.Item):
            cls.ensure_has_extension(obj, add_if_missing)
            return cast(PointcloudExtension[T], ItemPointcloudExtension(obj))
        elif isinstance(obj, pystac.Asset):
            if obj.owner is not None and not isinstance(obj.owner, pystac.Item):
                raise pystac.ExtensionTypeError(
                    "Pointcloud extension does not apply to Collection Assets."
                )
            cls.ensure_owner_has_extension(obj, add_if_missing)
            return cast(PointcloudExtension[T], AssetPointcloudExtension(obj))
        elif isinstance(obj, pystac.ItemAssetDefinition):
            cls.ensure_owner_has_extension(obj, add_if_missing)
            return cast(PointcloudExtension[T], ItemAssetsPointcloudExtension(obj))
        else:
            raise pystac.ExtensionTypeError(cls._ext_error_message(obj))

    @classmethod
    def summaries(
        cls, obj: pystac.Collection, add_if_missing: bool = False
    ) -> SummariesPointcloudExtension:
        cls.ensure_has_extension(obj, add_if_missing)
        return SummariesPointcloudExtension(obj)


class ItemPointcloudExtension(PointcloudExtension[pystac.Item]):
    """A concrete implementation of :class:`PointcloudExtension` on an
    :class:`~pystac.Item` that extends the properties of the Item to include
    properties defined in the :stac-ext:`Point Cloud Extension <pointcloud>`.

    This class should generally not be instantiated directly. Instead, call
    :meth:`PointcloudExtension.ext` on an :class:`~pystac.Item` to extend it.
    """

    item: pystac.Item
    properties: dict[str, Any]

    def __init__(self, item: pystac.Item):
        self.item = item
        self.properties = item.properties

    def __repr__(self) -> str:
        return f"<ItemPointcloudExtension Item id={self.item.id}>"


class AssetPointcloudExtension(PointcloudExtension[pystac.Asset]):
    """A concrete implementation of :class:`PointcloudExtension` on an
    :class:`~pystac.Asset` that extends the Asset fields to include properties defined
    in the :stac-ext:`Point Cloud Extension <pointcloud>`.

    This class should generally not be instantiated directly. Instead, call
    :meth:`PointcloudExtension.ext` on an :class:`~pystac.Asset` to extend it.
    """

    asset_href: str
    """The ``href`` value of the :class:`~pystac.Asset` being extended."""

    properties: dict[str, Any]
    """The :class:`~pystac.Asset` fields, including extension properties."""

    additional_read_properties: Iterable[dict[str, Any]] | None = None
    """If present, this will be a list containing 1 dictionary representing the
    properties of the owning :class:`~pystac.Item`."""

    def __init__(self, asset: pystac.Asset):
        self.asset_href = asset.href
        self.properties = asset.extra_fields
        if asset.owner and isinstance(asset.owner, pystac.Item):
            self.additional_read_properties = [asset.owner.properties]
            self.repr_id = f"href={asset.href} item.id={asset.owner.id}"
        else:
            self.repr_id = f"href={asset.href}"

    def __repr__(self) -> str:
        return f"<AssetPointcloudExtension Asset {self.repr_id}>"


class ItemAssetsPointcloudExtension(PointcloudExtension[pystac.ItemAssetDefinition]):
    properties: dict[str, Any]
    asset_defn: pystac.ItemAssetDefinition

    def __init__(self, item_asset: pystac.ItemAssetDefinition):
        self.asset_defn = item_asset
        self.properties = item_asset.properties


class SummariesPointcloudExtension(SummariesExtension):
    """A concrete implementation of :class:`~pystac.extensions.base.SummariesExtension`
    that extends the ``summaries`` field of a :class:`~pystac.Collection` to include
    properties defined in the :stac-ext:`Point Cloud Extension <pointcloud>`.
    """

    @property
    def count(self) -> RangeSummary[int] | None:
        return self.summaries.get_range(COUNT_PROP)

    @count.setter
    def count(self, v: RangeSummary[int] | None) -> None:
        self._set_summary(COUNT_PROP, v)

    @property
    def type(self) -> list[PhenomenologyType | str] | None:
        return self.summaries.get_list(TYPE_PROP)

    @type.setter
    def type(self, v: list[PhenomenologyType | str] | None) -> None:
        self._set_summary(TYPE_PROP, v)

    @property
    def encoding(self) -> list[str] | None:
        return self.summaries.get_list(ENCODING_PROP)

    @encoding.setter
    def encoding(self, v: list[str] | None) -> None:
        self._set_summary(ENCODING_PROP, v)

    @property
    def density(self) -> RangeSummary[float] | None:
        return self.summaries.get_range(DENSITY_PROP)

    @density.setter
    def density(self, v: RangeSummary[float] | None) -> None:
        self._set_summary(DENSITY_PROP, v)

    @property
    def statistics(self) -> list[Statistic] | None:
        return map_opt(
            lambda stats: [Statistic(d) for d in stats],
            self.summaries.get_list(STATISTICS_PROP),
        )

    @statistics.setter
    def statistics(self, v: list[Statistic] | None) -> None:
        self._set_summary(
            STATISTICS_PROP,
            map_opt(lambda stats: [s.to_dict() for s in stats], v),
        )


class PointcloudExtensionHooks(ExtensionHooks):
    schema_uri: str = SCHEMA_URI
    prev_extension_ids = {"pointcloud"}
    stac_object_types = {pystac.STACObjectType.ITEM}


POINTCLOUD_EXTENSION_HOOKS: ExtensionHooks = PointcloudExtensionHooks()



import json
import random
from string import ascii_letters

import pytest

import pystac
from pystac import ExtensionTypeError, Item
from pystac.collection import Collection
from pystac.extensions.storage import CloudPlatform, StorageExtension
from tests.utils import TestCases, assert_to_from_dict

NAIP_EXAMPLE_URI = TestCases.get_path("data-files/storage/item-naip.json")
NAIP_COLLECTION_URI = TestCases.get_path("data-files/storage/collection-naip.json")


@pytest.fixture
def naip_item() -> Item:
    return Item.from_file(NAIP_EXAMPLE_URI)


@pytest.fixture
def naip_collection() -> Collection:
    return Collection.from_file(NAIP_COLLECTION_URI)


def test_to_from_dict() -> None:
    with open(NAIP_EXAMPLE_URI) as f:
        item_dict = json.load(f)
    assert_to_from_dict(Item, item_dict)


def test_add_to(sample_item: Item) -> None:
    assert StorageExtension.get_schema_uri() not in sample_item.stac_extensions
    # Check that the URI gets added to stac_extensions
    StorageExtension.add_to(sample_item)
    assert StorageExtension.get_schema_uri() in sample_item.stac_extensions

    # Check that the URI only gets added once, regardless of how many times add_to
    # is called.
    StorageExtension.add_to(sample_item)
    StorageExtension.add_to(sample_item)

    eo_uris = [
        uri
        for uri in sample_item.stac_extensions
        if uri == StorageExtension.get_schema_uri()
    ]
    assert len(eo_uris) == 1


@pytest.mark.vcr()
def test_validate_storage(naip_item: Item) -> None:
    naip_item.validate()


def test_extend_invalid_object() -> None:
    link = pystac.Link("child", "https://some-domain.com/some/path/to.json")

    with pytest.raises(pystac.ExtensionTypeError):
        StorageExtension.ext(link)  # type: ignore


def test_extension_not_implemented(sample_item: Item) -> None:
    # Should raise exception if Item does not include extension URI
    with pytest.raises(pystac.ExtensionNotImplemented):
        _ = StorageExtension.ext(sample_item)

    # Should raise exception if owning Item does not include extension URI
    asset = sample_item.assets["thumbnail"]

    with pytest.raises(pystac.ExtensionNotImplemented):
        _ = StorageExtension.ext(asset)

    # Should succeed if Asset has no owner
    ownerless_asset = pystac.Asset.from_dict(asset.to_dict())
    _ = StorageExtension.ext(ownerless_asset)


def test_item_ext_add_to(sample_item: Item) -> None:
    assert StorageExtension.get_schema_uri() not in sample_item.stac_extensions

    _ = StorageExtension.ext(sample_item, add_if_missing=True)

    assert StorageExtension.get_schema_uri() in sample_item.stac_extensions


def test_asset_ext_add_to(sample_item: Item) -> None:
    assert StorageExtension.get_schema_uri() not in sample_item.stac_extensions
    asset = sample_item.assets["thumbnail"]

    _ = StorageExtension.ext(asset, add_if_missing=True)

    assert StorageExtension.get_schema_uri() in sample_item.stac_extensions


def test_asset_ext_add_to_ownerless_asset(sample_item: Item) -> None:
    asset_dict = sample_item.assets["thumbnail"].to_dict()
    asset = pystac.Asset.from_dict(asset_dict)

    with pytest.raises(pystac.STACError):
        _ = StorageExtension.ext(asset, add_if_missing=True)


def test_should_raise_exception_when_passing_invalid_extension_object() -> None:
    with pytest.raises(
        ExtensionTypeError, match=r"^StorageExtension does not apply to type 'object'$"
    ):
        # calling it wrong purposely so ---------v
        StorageExtension.ext(object())  # type: ignore


def test_summaries_platform(naip_collection: Collection) -> None:
    col_dict = naip_collection.to_dict()
    storage_summaries = StorageExtension.summaries(naip_collection)

    # Get
    assert storage_summaries.platform == col_dict["summaries"]["storage:platform"]
    # Set
    new_platform_summary = [random.choice([v for v in CloudPlatform])]
    assert storage_summaries.platform != new_platform_summary
    storage_summaries.platform = new_platform_summary
    assert storage_summaries.platform == new_platform_summary

    col_dict = naip_collection.to_dict()
    assert col_dict["summaries"]["storage:platform"] == new_platform_summary


def test_summaries_region(naip_collection: Collection) -> None:
    col_dict = naip_collection.to_dict()
    storage_summaries = StorageExtension.summaries(naip_collection)

    # Get
    assert storage_summaries.region == col_dict["summaries"]["storage:region"]
    # Set
    new_region_summary = [random.choice(ascii_letters)]
    assert storage_summaries.region != new_region_summary
    storage_summaries.region = new_region_summary
    assert storage_summaries.region == new_region_summary

    col_dict = naip_collection.to_dict()
    assert col_dict["summaries"]["storage:region"] == new_region_summary


def test_summaries_requester_pays(naip_collection: Collection) -> None:
    col_dict = naip_collection.to_dict()
    storage_summaries = StorageExtension.summaries(naip_collection)

    # Get
    assert (
        storage_summaries.requester_pays
        == col_dict["summaries"]["storage:requester_pays"]
    )

    # Set
    new_requester_pays_summary = [True]
    assert storage_summaries.requester_pays != new_requester_pays_summary
    storage_summaries.requester_pays = new_requester_pays_summary
    assert storage_summaries.requester_pays == new_requester_pays_summary

    col_dict = naip_collection.to_dict()
    assert col_dict["summaries"]["storage:requester_pays"] == new_requester_pays_summary


def test_summaries_tier(naip_collection: Collection) -> None:
    col_dict = naip_collection.to_dict()
    storage_summaries = StorageExtension.summaries(naip_collection)

    # Get
    assert storage_summaries.tier == col_dict["summaries"]["storage:tier"]

    # Set
    new_tier_summary = [random.choice(ascii_letters)]
    assert storage_summaries.tier != new_tier_summary
    storage_summaries.tier = new_tier_summary
    assert storage_summaries.tier == new_tier_summary

    col_dict = naip_collection.to_dict()
    assert col_dict["summaries"]["storage:tier"] == new_tier_summary


def test_summaries_adds_uri(naip_collection: Collection) -> None:
    naip_collection.stac_extensions = []
    with pytest.raises(
        pystac.ExtensionNotImplemented,
        match="Extension 'storage' is not implemented",
    ):
        StorageExtension.summaries(naip_collection, add_if_missing=False)

    _ = StorageExtension.summaries(naip_collection, add_if_missing=True)

    assert StorageExtension.get_schema_uri() in naip_collection.stac_extensions

    StorageExtension.remove_from(naip_collection)
    assert StorageExtension.get_schema_uri() not in naip_collection.stac_extensions


def test_item_apply(naip_item: Item) -> None:
    asset = random.choice(list(naip_item.assets.values()))

    storage_ext = StorageExtension.ext(asset)

    new_platform = random.choice(
        [v for v in CloudPlatform if v != storage_ext.platform]
    )
    new_region = random.choice(ascii_letters)
    new_requestor_pays = random.choice(
        [v for v in {True, False} if v != storage_ext.requester_pays]
    )
    new_tier = random.choice(ascii_letters)

    storage_ext.apply(
        platform=new_platform,
        region=new_region,
        requester_pays=new_requestor_pays,
        tier=new_tier,
    )

    assert storage_ext.platform == new_platform
    assert storage_ext.region == new_region
    assert storage_ext.requester_pays == new_requestor_pays
    assert storage_ext.tier == new_tier


@pytest.mark.vcr()
def test_asset_platform(naip_item: Item) -> None:
    # Grab a random asset with the platform property
    asset = random.choice(
        [
            _asset
            for _asset in naip_item.assets.values()
            if "storage:platform" in _asset.to_dict()
        ]
    )

    storage_ext = StorageExtension.ext(asset)

    # Get
    assert storage_ext.platform == asset.extra_fields.get("storage:platform")

    # Set
    new_platform = random.choice(
        [val for val in CloudPlatform if val != storage_ext.platform]
    )
    storage_ext.platform = new_platform
    assert storage_ext.platform == new_platform

    naip_item.validate()


@pytest.mark.vcr()
def test_asset_region(naip_item: Item) -> None:
    # Grab a random asset with the platform property
    asset = random.choice(
        [
            _asset
            for _asset in naip_item.assets.values()
            if "storage:region" in _asset.to_dict()
        ]
    )

    storage_ext = StorageExtension.ext(asset)

    # Get
    assert storage_ext.region == asset.extra_fields.get("storage:region")

    # Set
    new_region = random.choice(
        [val for val in CloudPlatform if val != storage_ext.region]
    )
    storage_ext.region = new_region
    assert storage_ext.region == new_region

    naip_item.validate()

    # Set to None
    storage_ext.region = None
    assert "storage:region" not in asset.extra_fields


@pytest.mark.vcr()
def test_asset_requester_pays(naip_item: Item) -> None:
    # Grab a random asset with the platform property
    asset = random.choice(
        [
            _asset
            for _asset in naip_item.assets.values()
            if "storage:requester_pays" in _asset.to_dict()
        ]
    )

    storage_ext = StorageExtension.ext(asset)

    # Get
    assert storage_ext.requester_pays == asset.extra_fields.get(
        "storage:requester_pays"
    )

    # Set
    new_requester_pays = True if not storage_ext.requester_pays else False
    storage_ext.requester_pays = new_requester_pays
    assert storage_ext.requester_pays == new_requester_pays

    naip_item.validate()

    # Set to None
    storage_ext.requester_pays = None
    assert "storage:requester_pays" not in asset.extra_fields


@pytest.mark.vcr()
def test_asset_tier(naip_item: Item) -> None:
    # Grab a random asset with the platform property
    asset = random.choice(
        [
            _asset
            for _asset in naip_item.assets.values()
            if "storage:tier" in _asset.to_dict()
        ]
    )

    storage_ext = StorageExtension.ext(asset)

    # Get
    assert storage_ext.tier == asset.extra_fields.get("storage:tier")

    # Set
    new_tier = random.choice(ascii_letters)
    storage_ext.tier = new_tier
    assert storage_ext.tier == new_tier

    naip_item.validate()

    # Set to None
    storage_ext.tier = None
    assert "storage:tier" not in asset.extra_fields



"""
Script to download the examples from the stac-spec repository.
This is used when upgrading to a new version of STAC.
"""

import argparse
import json
import os
import tempfile
from subprocess import call
from typing import Any
from urllib.error import HTTPError

import pystac
from pystac.serialization import identify_stac_object


def remove_bad_collection(js: dict[str, Any]) -> dict[str, Any]:
    links: list[dict[str, Any]] | None = js.get("links")
    if links is not None:
        filtered_links: list[dict[str, Any]] = []
        for link in links:
            rel = link.get("rel")
            if rel is not None and rel == "collection":
                href: str = link["href"]
                try:
                    json.loads(pystac.StacIO.default().read_text(href))
                    filtered_links.append(link)
                except (HTTPError, FileNotFoundError, json.decoder.JSONDecodeError):
                    print(f"===REMOVING UNREADABLE COLLECTION AT {href}")
            else:
                filtered_links.append(link)
        js["links"] = filtered_links
    return js


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Get examples from the stac-spec repo."
    )
    parser.add_argument(
        "previous_version",
        metavar="PREVIOUS_VERSION",
        help="The previous STAC_VERSION that examples have already been pulled from.",
    )

    args = parser.parse_args()

    stac_repo = "https://github.com/radiantearth/stac-spec"
    stac_spec_tag = f"v{pystac.get_stac_version()}"

    examples_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), "examples"))

    with tempfile.TemporaryDirectory() as tmp_dir:
        call(
            [
                "git",
                "clone",
                "--depth",
                "1",
                "--branch",
                stac_spec_tag,
                stac_repo,
                tmp_dir,
            ]
        )

        example_dirs: list[str] = []
        for root, _, _ in os.walk(tmp_dir):
            example_dirs.append(os.path.join(root))

        example_csv_lines = set()

        for example_dir in example_dirs:
            for root, _, files in os.walk(example_dir):
                for fname in files:
                    if fname.endswith(".json"):
                        path = os.path.join(root, fname)
                        with open(path) as f:
                            try:
                                js: dict[str, Any] = json.loads(f.read())
                            except json.decoder.JSONDecodeError:
                                # Account for bad examples that can't be parsed.
                                js = {}
                            example_version = js.get("stac_version")
                        if (
                            example_version is not None
                            and example_version > args.previous_version
                        ):
                            relpath = "{}/{}".format(
                                pystac.get_stac_version(),
                                path.replace(f"{tmp_dir}/", ""),
                            )
                            target_path = os.path.join(examples_dir, relpath)

                            print(f"Creating example at {target_path}")

                            info = identify_stac_object(js)

                            # Handle the case where there are collection links that
                            # don't exist.
                            if info.object_type == pystac.STACObjectType.ITEM:
                                js = remove_bad_collection(js)

                            d = os.path.dirname(target_path)
                            if not os.path.isdir(d):
                                os.makedirs(d)

                            with open(target_path, "w") as f:
                                f.write(json.dumps(js, indent=4))

                            # Add info to the new example-info.csv lines
                            line_info: list[str] = [
                                relpath,
                                info.object_type,
                                example_version,
                                "|".join(info.extensions),
                            ]
                            line = '"{}"'.format('","'.join(line_info))
                            example_csv_lines.add(line)

        # Write the new example-info.csv lines into a temp file for inspection
        with open(os.path.join(examples_dir, "examples-info-NEW.csv"), "w") as f:
            txt = "\n".join(sorted(example_csv_lines))
            f.write(txt)



"""
Script to change the version property of the test files for PySTAC.
This is used when upgrading to a new version of STAC.
"""

import argparse
import json
import os
import re

import pystac

TARGET_VERSION = pystac.get_stac_version()


def migrate(path: str) -> None:
    try:
        with open(path) as f:
            stac_json = json.load(f)
    except json.decoder.JSONDecodeError:
        print(f"Cannot read {path}")
        raise

    if "stac_version" in stac_json:
        cur_ver = stac_json["stac_version"]
        if not cur_ver == TARGET_VERSION:
            print(
                "  - Migrating {} from {} to {}...".format(
                    path, cur_ver, TARGET_VERSION
                )
            )
            obj = pystac.read_dict(stac_json, href=path)
            migrated = obj.to_dict(include_self_link=False)
            with open(path, "w") as f:
                json.dump(migrated, f, indent=2)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Process some integers.")
    parser.add_argument(
        "--file", metavar="FILE", help="Only migrate this specific file."
    )

    args = parser.parse_args()

    if args.file:
        migrate(os.path.abspath(args.file))
    else:
        data_files_dir = os.path.dirname(os.path.realpath(__file__))

        # Skip examples directory, which contains version specific STACs...
        dirs_to_check = [
            os.path.join(data_files_dir, x)
            for x in os.listdir(data_files_dir)
            if x != "examples"
        ]

        for d in dirs_to_check:
            print(f"checking in {d}..")
            for root, _, files in os.walk(d):
                # Skip directories with a version tag
                if re.match(r".*-v\d.*", root):
                    continue
                for fname in files:
                    if fname.endswith(".json"):
                        path = os.path.join(root, fname)
                        migrate(path)



__all__ = [
    "merge_common_properties",
    "migrate_to_latest",
    "STACVersionRange",
    "identify_stac_object",
    "identify_stac_object_type",
]
from pystac.serialization.common_properties import merge_common_properties
from pystac.serialization.identify import (
    STACVersionRange,
    identify_stac_object,
    identify_stac_object_type,
)
from pystac.serialization.migrate import migrate_to_latest



import json
from copy import deepcopy
from typing import Any

import pytest

import pystac
from pystac import ExtensionTypeError, Item
from pystac.errors import ExtensionNotImplemented
from pystac.extensions.projection import ProjectionExtension
from pystac.utils import get_opt
from tests.utils import TestCases, assert_to_from_dict

WKT2 = """
GEOGCS["WGS 84",
    DATUM["WGS_1984",
        SPHEROID["WGS 84",6378137,298.257223563,
            AUTHORITY["EPSG","7030"]],
        AUTHORITY["EPSG","6326"]],
    PRIMEM["Greenwich",0,
        AUTHORITY["EPSG","8901"]],
    UNIT["degree",0.0174532925199433,
        AUTHORITY["EPSG","9122"]],
    AXIS["Latitude",NORTH],
    AXIS["Longitude",EAST],
    AUTHORITY["EPSG","4326"]]
"""
PROJJSON = json.loads(
    """
{
    "$schema": "https://proj.org/schemas/v0.1/projjson.schema.json",
    "type": "GeographicCRS",
    "name": "WGS 84",
    "datum": {
        "type": "GeodeticReferenceFrame",
        "name": "World Geodetic System 1984",
        "ellipsoid": {
            "name": "WGS 84",
            "semi_major_axis": 6378137,
            "inverse_flattening": 298.257223563
        }
    },
    "coordinate_system": {
        "subtype": "ellipsoidal",
        "axis": [
        {
            "name": "Geodetic latitude",
            "abbreviation": "Lat",
            "direction": "north",
            "unit": "degree"
        },
        {
            "name": "Geodetic longitude",
            "abbreviation": "Lon",
            "direction": "east",
            "unit": "degree"
        }
        ]
    },
    "area": "World",
    "bbox": {
        "south_latitude": -90,
        "west_longitude": -180,
        "north_latitude": 90,
        "east_longitude": 180
    },
    "id": {
        "authority": "EPSG",
        "code": 4326
    }
}
"""
)


@pytest.fixture
def example_uri() -> str:
    return TestCases.get_path("data-files/projection/example-landsat8.json")


@pytest.fixture
def proj_item(example_uri: str) -> Item:
    return pystac.Item.from_file(example_uri)


@pytest.fixture
def example_summaries_uri() -> str:
    return TestCases.get_path("data-files/projection/collection-with-summaries.json")


def test_to_from_dict(example_uri: str) -> None:
    with open(example_uri) as f:
        d = json.load(f)
    assert_to_from_dict(pystac.Item, d)


def test_apply() -> None:
    item = next(TestCases.case_2().get_items(recursive=True))
    assert not ProjectionExtension.has_extension(item)

    ProjectionExtension.add_to(item)
    ProjectionExtension.ext(item).apply(
        epsg=4326,
        wkt2=WKT2,
        projjson=PROJJSON,
        geometry=item.geometry,
        bbox=item.bbox,
        centroid={"lat": 0.0, "lon": 1.0},
        shape=[100, 100],
        transform=[30.0, 0.0, 224985.0, 0.0, -30.0, 6790215.0, 0.0, 0.0, 1.0],
    )


@pytest.mark.vcr()
def test_partial_apply(proj_item: Item) -> None:
    ProjectionExtension.ext(proj_item).apply(epsg=1111)

    assert ProjectionExtension.ext(proj_item).epsg == 1111
    proj_item.validate()


@pytest.mark.vcr()
def test_validate_proj(proj_item: Item) -> None:
    proj_item.validate()


@pytest.mark.vcr()
def test_epsg(proj_item: Item) -> None:
    # Get
    assert "proj:epsg" not in proj_item.properties
    assert "proj:code" in proj_item.properties
    proj_epsg = ProjectionExtension.ext(proj_item).epsg
    assert f"EPSG:{proj_epsg}" == proj_item.properties["proj:code"]

    # Set
    assert proj_epsg is not None
    ProjectionExtension.ext(proj_item).epsg = proj_epsg + 100
    assert f"EPSG:{proj_epsg + 100}" == proj_item.properties["proj:code"]

    # Get from Asset
    asset_no_prop = proj_item.assets["B1"]
    asset_prop = proj_item.assets["B8"]
    assert (
        ProjectionExtension.ext(asset_no_prop).epsg
        == ProjectionExtension.ext(proj_item).epsg
    )
    assert ProjectionExtension.ext(asset_prop).epsg == 9999

    # Set to Asset
    ProjectionExtension.ext(asset_no_prop).epsg = 8888
    assert (
        ProjectionExtension.ext(asset_no_prop).epsg
        != ProjectionExtension.ext(proj_item).epsg
    )
    assert ProjectionExtension.ext(asset_no_prop).epsg == 8888

    # Validate
    proj_item.validate()


def test_optional_epsg() -> None:
    example_uri = TestCases.get_path("data-files/projection/optional-epsg.json")
    proj_item = pystac.Item.from_file(example_uri)

    # No proj info on item
    assert "proj:epsg" not in proj_item.properties
    assert "proj:code" not in proj_item.properties

    # Some proj info on assets
    asset_no_prop = proj_item.assets["metadata"]
    assert ProjectionExtension.ext(asset_no_prop).epsg is None

    asset_prop = proj_item.assets["visual"]
    assert ProjectionExtension.ext(asset_prop).epsg == 32618


@pytest.mark.vcr()
def test_wkt2(proj_item: Item) -> None:
    # Get
    assert "proj:wkt2" in proj_item.properties
    proj_wkt2 = ProjectionExtension.ext(proj_item).wkt2
    assert proj_wkt2 == proj_item.properties["proj:wkt2"]

    # Set
    ProjectionExtension.ext(proj_item).wkt2 = WKT2
    assert WKT2 == proj_item.properties["proj:wkt2"]

    # Get from Asset
    asset_no_prop = proj_item.assets["B1"]
    asset_prop = proj_item.assets["B8"]
    assert (
        ProjectionExtension.ext(asset_no_prop).wkt2
        == ProjectionExtension.ext(proj_item).wkt2
    )
    assert "TEST_TEXT" in get_opt(ProjectionExtension.ext(asset_prop).wkt2)

    # Set to Asset
    asset_value = "TEST TEXT 2"
    ProjectionExtension.ext(asset_no_prop).wkt2 = asset_value
    assert (
        ProjectionExtension.ext(asset_no_prop).wkt2
        != ProjectionExtension.ext(proj_item).wkt2
    )
    assert ProjectionExtension.ext(asset_no_prop).wkt2 == asset_value

    # Validate
    proj_item.validate()


@pytest.mark.vcr()
def test_projjson(proj_item: Item) -> None:
    # Get
    assert "proj:projjson" in proj_item.properties
    proj_projjson = ProjectionExtension.ext(proj_item).projjson
    assert proj_projjson == proj_item.properties["proj:projjson"]

    # Set
    ProjectionExtension.ext(proj_item).projjson = PROJJSON
    assert PROJJSON == proj_item.properties["proj:projjson"]

    # Get from Asset
    asset_no_prop = proj_item.assets["B1"]
    asset_prop = proj_item.assets["B8"]
    assert (
        ProjectionExtension.ext(asset_no_prop).projjson
        == ProjectionExtension.ext(proj_item).projjson
    )
    asset_prop_json = ProjectionExtension.ext(asset_prop).projjson
    assert asset_prop_json is not None
    assert asset_prop_json["id"]["code"] == 9999

    # Set to Asset
    asset_value = deepcopy(PROJJSON)
    asset_value["id"]["code"] = 7777
    ProjectionExtension.ext(asset_no_prop).projjson = asset_value
    assert (
        ProjectionExtension.ext(asset_no_prop).projjson
        != ProjectionExtension.ext(proj_item).projjson
    )
    asset_no_prop_json = ProjectionExtension.ext(asset_no_prop).projjson
    assert asset_no_prop_json is not None
    assert asset_no_prop_json["id"]["code"] == 7777

    # Validate
    proj_item.validate()

    # Ensure setting bad projjson fails validation
    with pytest.raises(pystac.STACValidationError):
        ProjectionExtension.ext(proj_item).projjson = {"bad": "data"}
        proj_item.validate()


def test_crs_string(proj_item: Item) -> None:
    ProjectionExtension.remove_from(proj_item)
    for key in list(proj_item.properties.keys()):
        if key.startswith("proj:"):
            proj_item.properties.pop(key)
    assert proj_item.properties.get("proj:code") is None
    assert proj_item.properties.get("proj:epsg") is None
    assert proj_item.properties.get("proj:wkt2") is None
    assert proj_item.properties.get("proj:projjson") is None

    projection = ProjectionExtension.ext(proj_item, add_if_missing=True)
    assert projection.crs_string is None

    projection.projjson = PROJJSON
    assert projection.crs_string == json.dumps(PROJJSON)

    projection.wkt2 = WKT2
    assert projection.crs_string == WKT2

    projection.epsg = 4326
    assert projection.crs_string == "EPSG:4326"

    projection.code = "IAU_2015:49900"
    assert projection.crs_string == "IAU_2015:49900"


@pytest.mark.vcr()
def test_geometry(proj_item: Item) -> None:
    # Get
    assert "proj:geometry" in proj_item.properties
    proj_geometry = ProjectionExtension.ext(proj_item).geometry
    assert proj_geometry == proj_item.properties["proj:geometry"]

    # Set
    ProjectionExtension.ext(proj_item).geometry = proj_item.geometry
    assert proj_item.geometry == proj_item.properties["proj:geometry"]

    # Get from Asset
    asset_no_prop = proj_item.assets["B1"]
    asset_prop = proj_item.assets["B8"]
    assert (
        ProjectionExtension.ext(asset_no_prop).geometry
        == ProjectionExtension.ext(proj_item).geometry
    )
    asset_prop_geometry = ProjectionExtension.ext(asset_prop).geometry
    assert asset_prop_geometry is not None
    assert asset_prop_geometry["coordinates"][0][0], [0.0 == 0.0]

    # Set to Asset
    asset_value: dict[str, Any] = {"type": "Point", "coordinates": [1.0, 2.0]}
    ProjectionExtension.ext(asset_no_prop).geometry = asset_value
    assert (
        ProjectionExtension.ext(asset_no_prop).geometry
        != ProjectionExtension.ext(proj_item).geometry
    )
    assert ProjectionExtension.ext(asset_no_prop).geometry == asset_value

    # Validate
    proj_item.validate()

    # Ensure setting bad geometry fails validation
    with pytest.raises(pystac.STACValidationError):
        ProjectionExtension.ext(proj_item).geometry = {"bad": "data"}
        proj_item.validate()


@pytest.mark.vcr()
def test_bbox(proj_item: Item) -> None:
    # Get
    assert "proj:bbox" in proj_item.properties
    proj_bbox = ProjectionExtension.ext(proj_item).bbox
    assert proj_bbox == proj_item.properties["proj:bbox"]

    # Set
    ProjectionExtension.ext(proj_item).bbox = [1.0, 2.0, 3.0, 4.0]
    assert proj_item.properties["proj:bbox"] == [1.0, 2.0, 3.0, 4.0]

    # Get from Asset
    asset_no_prop = proj_item.assets["B1"]
    asset_prop = proj_item.assets["B8"]
    assert (
        ProjectionExtension.ext(asset_no_prop).bbox
        == ProjectionExtension.ext(proj_item).bbox
    )
    assert ProjectionExtension.ext(asset_prop).bbox == [1.0, 2.0, 3.0, 4.0]

    # Set to Asset
    asset_value = [10.0, 20.0, 30.0, 40.0]
    ProjectionExtension.ext(asset_no_prop).bbox = asset_value
    assert (
        ProjectionExtension.ext(asset_no_prop).bbox
        != ProjectionExtension.ext(proj_item).bbox
    )
    assert ProjectionExtension.ext(asset_no_prop).bbox == asset_value

    # Validate
    proj_item.validate()


@pytest.mark.vcr()
def test_centroid(proj_item: Item) -> None:
    # Get
    assert "proj:centroid" in proj_item.properties
    proj_centroid = ProjectionExtension.ext(proj_item).centroid
    assert proj_centroid == proj_item.properties["proj:centroid"]

    # Set
    new_val = {"lat": 2.0, "lon": 3.0}
    ProjectionExtension.ext(proj_item).centroid = new_val
    assert proj_item.properties["proj:centroid"] == new_val

    # Get from Asset
    asset_no_prop = proj_item.assets["B1"]
    asset_prop = proj_item.assets["B8"]
    assert (
        ProjectionExtension.ext(asset_no_prop).centroid
        == ProjectionExtension.ext(proj_item).centroid
    )
    assert ProjectionExtension.ext(asset_prop).centroid == {"lat": 0.5, "lon": 0.3}

    # Set to Asset
    asset_value = {"lat": 1.5, "lon": 1.3}
    ProjectionExtension.ext(asset_no_prop).centroid = asset_value
    assert (
        ProjectionExtension.ext(asset_no_prop).centroid
        != ProjectionExtension.ext(proj_item).centroid
    )
    assert ProjectionExtension.ext(asset_no_prop).centroid == asset_value

    # Validate
    proj_item.validate()

    # Ensure setting bad centroid fails validation
    with pytest.raises(pystac.STACValidationError):
        ProjectionExtension.ext(proj_item).centroid = {"lat": 2.0, "lng": 3.0}
        proj_item.validate()


@pytest.mark.vcr()
def test_shape(proj_item: Item) -> None:
    # Get
    assert "proj:shape" in proj_item.properties
    proj_shape = ProjectionExtension.ext(proj_item).shape
    assert proj_shape == proj_item.properties["proj:shape"]

    # Set
    new_val = [100, 200]
    ProjectionExtension.ext(proj_item).shape = new_val
    assert proj_item.properties["proj:shape"] == new_val

    # Get from Asset
    asset_no_prop = proj_item.assets["B1"]
    asset_prop = proj_item.assets["B8"]
    assert (
        ProjectionExtension.ext(asset_no_prop).shape
        == ProjectionExtension.ext(proj_item).shape
    )
    assert ProjectionExtension.ext(asset_prop).shape == [16781, 16621]

    # Set to Asset
    asset_value = [1, 2]
    ProjectionExtension.ext(asset_no_prop).shape = asset_value
    assert (
        ProjectionExtension.ext(asset_no_prop).shape
        != ProjectionExtension.ext(proj_item).shape
    )
    assert ProjectionExtension.ext(asset_no_prop).shape == asset_value

    # Validate
    proj_item.validate()


@pytest.mark.vcr()
def test_transform(proj_item: Item) -> None:
    # Get
    assert "proj:transform" in proj_item.properties
    proj_transform = ProjectionExtension.ext(proj_item).transform
    assert proj_transform == proj_item.properties["proj:transform"]

    # Set
    new_val = [1.0, 2.0, 3.0, 4.0, 5.0, 6.0]
    ProjectionExtension.ext(proj_item).transform = new_val
    assert proj_item.properties["proj:transform"] == new_val

    # Get from Asset
    asset_no_prop = proj_item.assets["B1"]
    asset_prop = proj_item.assets["B8"]
    assert (
        ProjectionExtension.ext(asset_no_prop).transform
        == ProjectionExtension.ext(proj_item).transform
    )
    assert ProjectionExtension.ext(asset_prop).transform == [
        15.0,
        0.0,
        224992.5,
        0.0,
        -15.0,
        6790207.5,
        0.0,
        0.0,
        1.0,
    ]

    # Set to Asset
    asset_value = [2.0, 4.0, 6.0, 8.0, 10.0, 12.0]
    ProjectionExtension.ext(asset_no_prop).transform = asset_value
    assert (
        ProjectionExtension.ext(asset_no_prop).transform
        != ProjectionExtension.ext(proj_item).transform
    )
    assert ProjectionExtension.ext(asset_no_prop).transform == asset_value

    # Validate
    proj_item.validate()


def test_extension_not_implemented(proj_item: Item) -> None:
    # Should raise exception if Item does not include extension URI
    proj_item.stac_extensions.remove(ProjectionExtension.get_schema_uri())

    with pytest.raises(pystac.ExtensionNotImplemented):
        _ = ProjectionExtension.ext(proj_item)

    # Should raise exception if owning Item does not include extension URI
    asset = proj_item.assets["B8"]

    with pytest.raises(pystac.ExtensionNotImplemented):
        _ = ProjectionExtension.ext(asset)

    # Should succeed if Asset has no owner
    ownerless_asset = pystac.Asset.from_dict(asset.to_dict())
    _ = ProjectionExtension.ext(ownerless_asset)


def test_item_ext_add_to(proj_item: Item) -> None:
    proj_item.stac_extensions.remove(ProjectionExtension.get_schema_uri())
    assert ProjectionExtension.get_schema_uri() not in proj_item.stac_extensions

    _ = ProjectionExtension.ext(proj_item, add_if_missing=True)

    assert ProjectionExtension.get_schema_uri() in proj_item.stac_extensions


def test_asset_ext_add_to(proj_item: Item) -> None:
    proj_item.stac_extensions.remove(ProjectionExtension.get_schema_uri())
    assert ProjectionExtension.get_schema_uri() not in proj_item.stac_extensions
    asset = proj_item.assets["B8"]

    _ = ProjectionExtension.ext(asset, add_if_missing=True)

    assert ProjectionExtension.get_schema_uri() in proj_item.stac_extensions


def test_should_raise_exception_when_passing_invalid_extension_object() -> None:
    with pytest.raises(
        ExtensionTypeError,
        match=r"^ProjectionExtension does not apply to type 'object'$",
    ):
        # intentionally calling this wrong so ---vvv
        ProjectionExtension.ext(object())  # type: ignore


def test_get_summaries(example_summaries_uri: str) -> None:
    col = pystac.Collection.from_file(example_summaries_uri)
    proj_summaries = ProjectionExtension.summaries(col)

    # Get

    epsg_summaries = proj_summaries.epsg
    assert epsg_summaries is not None
    assert epsg_summaries == [32614]


def test_set_summaries(example_summaries_uri: str) -> None:
    col = pystac.Collection.from_file(example_summaries_uri)
    proj_summaries = ProjectionExtension.summaries(col)

    # Set

    proj_summaries.epsg = [4326]

    col_dict = col.to_dict()
    assert col_dict["summaries"]["proj:code"] == ["EPSG:4326"]


def test_summaries_adds_uri(example_summaries_uri: str) -> None:
    col = pystac.Collection.from_file(example_summaries_uri)
    col.stac_extensions = []
    with pytest.raises(
        pystac.ExtensionNotImplemented, match="Extension 'proj' is not implemented"
    ):
        ProjectionExtension.summaries(col, add_if_missing=False)

    ProjectionExtension.summaries(col, True)

    assert ProjectionExtension.get_schema_uri() in col.stac_extensions

    ProjectionExtension.remove_from(col)
    assert ProjectionExtension.get_schema_uri() not in col.stac_extensions


def test_no_args_for_extension_class(item: Item) -> None:
    item.ext.add("proj")
    with pytest.raises(TypeError, match="takes 1 positional argument but 2 were given"):
        item.ext.proj.apply(32614)  # type:ignore


def test_set_both_code_and_epsg(item: Item) -> None:
    item.ext.add("proj")
    with pytest.raises(KeyError, match="Only one of the options"):
        item.ext.proj.apply(epsg=32614, code="EPSG:32614")


@pytest.mark.vcr()
def test_get_set_code(projection_landsat8_item: Item) -> None:
    proj_item = projection_landsat8_item
    assert proj_item.ext.proj.code == proj_item.properties["proj:code"]
    assert proj_item.ext.proj.epsg == 32614

    proj_item.ext.proj.code = "IAU_2015:30100"
    assert proj_item.ext.proj.epsg is None
    assert proj_item.properties["proj:code"] == "IAU_2015:30100"


def test_migrate_item_on_1_2() -> None:
    old = "https://stac-extensions.github.io/projection/v1.2.0/schema.json"
    current = "https://stac-extensions.github.io/projection/v2.0.0/schema.json"

    path = TestCases.get_path("data-files/projection/example-with-version-1.2.json")
    with pytest.warns(UserWarning, match="surprising behavior"):
        item = Item.from_file(path)

    assert old not in item.stac_extensions
    assert current in item.stac_extensions

    assert item.ext.proj.epsg == 32613
    assert item.ext.proj.code == "EPSG:32613"

    assert item.assets["B1"].ext.proj.epsg == 32613
    assert item.assets["B1"].ext.proj.code == "EPSG:32613"

    assert item.assets["B8"].ext.proj.epsg == 9998
    assert item.assets["B8"].ext.proj.code == "EPSG:9998"


def test_migrate_item_on_1_1() -> None:
    old = "https://stac-extensions.github.io/projection/v1.1.0/schema.json"
    current = "https://stac-extensions.github.io/projection/v2.0.0/schema.json"

    path = TestCases.get_path("data-files/projection/example-with-version-1.1.json")
    item = Item.from_file(path)

    assert old not in item.stac_extensions
    assert current in item.stac_extensions

    assert item.ext.proj.epsg == 32614
    assert item.ext.proj.code == "EPSG:32614"

    assert item.assets["B1"].ext.proj.epsg == 32614
    assert item.assets["B1"].ext.proj.code == "EPSG:32614"

    assert item.assets["B8"].ext.proj.epsg == 9999
    assert item.assets["B8"].ext.proj.code == "EPSG:9999"


def test_migrate_collection_item_assets() -> None:
    old = "https://stac-extensions.github.io/projection/v1.1.0/schema.json"
    current = "https://stac-extensions.github.io/projection/v2.0.0/schema.json"

    path = TestCases.get_path("data-files/projection/collection-with-summaries.json")
    collection = pystac.Collection.from_file(path)

    assert old not in collection.stac_extensions
    assert current in collection.stac_extensions

    for item_asset in collection.item_assets.values():
        assert item_asset.ext.proj.epsg == 32659
        assert item_asset.ext.proj.code == "EPSG:32659"


def test_older_extension_version(projection_landsat8_item: Item) -> None:
    old = "https://stac-extensions.github.io/projection/v1.0.0/schema.json"
    current = "https://stac-extensions.github.io/projection/v2.0.0/schema.json"

    stac_extensions = set(projection_landsat8_item.stac_extensions)
    stac_extensions.remove(current)
    stac_extensions.add(old)
    item_as_dict = projection_landsat8_item.to_dict(
        include_self_link=False, transform_hrefs=False
    )
    item_as_dict["stac_extensions"] = list(stac_extensions)
    item = Item.from_dict(item_as_dict, migrate=False)
    assert ProjectionExtension.has_extension(item)
    assert old in item.stac_extensions

    migrated_item = pystac.Item.from_dict(item_as_dict, migrate=True)
    assert ProjectionExtension.has_extension(migrated_item)
    assert current in migrated_item.stac_extensions


def test_newer_extension_version(projection_landsat8_item: Item) -> None:
    new = "https://stac-extensions.github.io/projection/v2.1.0/schema.json"
    current = "https://stac-extensions.github.io/projection/v2.0.0/schema.json"

    stac_extensions = set(projection_landsat8_item.stac_extensions)
    stac_extensions.remove(current)
    stac_extensions.add(new)
    item_as_dict = projection_landsat8_item.to_dict(
        include_self_link=False, transform_hrefs=False
    )
    item_as_dict["stac_extensions"] = list(stac_extensions)
    item = Item.from_dict(item_as_dict)
    assert ProjectionExtension.has_extension(item)
    assert new in item.stac_extensions

    migrated_item = pystac.Item.from_dict(item_as_dict, migrate=True)
    assert ProjectionExtension.has_extension(migrated_item)
    assert new in migrated_item.stac_extensions


def test_ext_syntax(projection_landsat8_item: pystac.Item) -> None:
    ext_item = projection_landsat8_item
    assert ext_item.ext.proj.epsg == 32614
    assert ext_item.assets["B1"].ext.proj.epsg == 32614


def test_ext_syntax_remove(projection_landsat8_item: pystac.Item) -> None:
    ext_item = projection_landsat8_item
    ext_item.ext.remove("proj")
    with pytest.raises(ExtensionNotImplemented):
        ext_item.ext.proj


def test_ext_syntax_add(item: pystac.Item) -> None:
    item.ext.add("proj")
    assert isinstance(item.ext.proj, ProjectionExtension)


def test_v1_from_dict() -> None:
    with open(
        TestCases.get_path("data-files/projection/example-with-version-1.1.json")
    ) as f:
        data = json.load(f)
    item = pystac.Item.from_dict(data, migrate=False)
    assert item.ext.proj.epsg is not None
    assert item.ext.proj.crs_string is not None


def test_v1_crs_string() -> None:
    with open(TestCases.get_path("data-files/projection/another-1.1.json")) as f:
        data = json.load(f)
    item = pystac.Item.from_dict(data, migrate=False)
    assert item.ext.proj.epsg is not None
    assert item.ext.proj.crs_string == "EPSG:32617"


def test_none_epsg(item: Item) -> None:
    # https://github.com/stac-utils/pystac/issues/1543
    d = item.to_dict(include_self_link=False, transform_hrefs=False)
    d["stac_version"] = "1.0.0"
    d["stac_extensions"] = [
        "https://stac-extensions.github.io/projection/v1.1.0/schema.json"
    ]
    d["properties"]["proj:epsg"] = None
    item = Item.from_dict(d, migrate=True)
    assert item.ext.proj.code is None



import csv
import os
from datetime import datetime, timezone
from typing import Any

import pystac
from pystac import (
    Asset,
    Catalog,
    Collection,
    Extent,
    Item,
    MediaType,
    SpatialExtent,
    TemporalExtent,
)

TEST_LABEL_CATALOG = {
    "country-1": {
        "area-1-1": {
            "dsm": "area-1-1_dsm.tif",
            "ortho": "area-1-1_ortho.tif",
            "labels": "area-1-1_labels.geojson",
        },
        "area-1-2": {
            "dsm": "area-1-2_dsm.tif",
            "ortho": "area-1-2_ortho.tif",
            "labels": "area-1-2_labels.geojson",
        },
    },
    "country-2": {
        "area-2-1": {
            "dsm": "area-2-1_dsm.tif",
            "ortho": "area-2-1_ortho.tif",
            "labels": "area-2-1_labels.geojson",
        },
        "area-2-2": {
            "dsm": "area-2-2_dsm.tif",
            "ortho": "area-2-2_ortho.tif",
            "labels": "area-2-2_labels.geojson",
        },
    },
}

ARBITRARY_GEOM: dict[str, Any] = {
    "type": "Polygon",
    "coordinates": [
        [
            [-2.5048828125, 3.8916575492899987],
            [-1.9610595703125, 3.8916575492899987],
            [-1.9610595703125, 4.275202171119132],
            [-2.5048828125, 4.275202171119132],
            [-2.5048828125, 3.8916575492899987],
        ]
    ],
}

ARBITRARY_BBOX: list[float] = [
    ARBITRARY_GEOM["coordinates"][0][0][0],
    ARBITRARY_GEOM["coordinates"][0][0][1],
    ARBITRARY_GEOM["coordinates"][0][1][0],
    ARBITRARY_GEOM["coordinates"][0][1][1],
]

ARBITRARY_EXTENT = Extent(
    spatial=SpatialExtent.from_coordinates(ARBITRARY_GEOM["coordinates"]),
    temporal=TemporalExtent.from_now(),
)


class ExampleInfo:
    def __init__(
        self,
        path: str,
        object_type: pystac.STACObjectType,
        stac_version: str,
        extensions: list[str],
        valid: bool,
    ) -> None:
        self.path = path
        self.object_type = object_type
        self.stac_version = stac_version
        self.extensions = extensions
        self.valid = valid


class TestCases:
    bad_catalog_case = "data-files/catalogs/invalid-catalog/catalog.json"

    @staticmethod
    def get_path(rel_path: str) -> str:
        return os.path.abspath(os.path.join(os.path.dirname(__file__), "..", rel_path))

    @staticmethod
    def get_examples_info() -> list[ExampleInfo]:
        examples: list[ExampleInfo] = []

        info_path = TestCases.get_path("data-files/examples/example-info.csv")
        with open(TestCases.get_path("data-files/examples/example-info.csv")) as f:
            for row in csv.reader(f):
                path = os.path.abspath(os.path.join(os.path.dirname(info_path), row[0]))
                object_type = row[1]
                stac_version = row[2]
                extensions: list[str] = []
                if row[3]:
                    extensions = row[3].split("|")

                valid = True
                if len(row) > 4:
                    # The 5th column will be "INVALID" if the example
                    # shouldn't pass validation
                    valid = row[4] != "INVALID"

                examples.append(
                    ExampleInfo(
                        path=path,
                        object_type=pystac.STACObjectType(object_type),
                        stac_version=stac_version,
                        extensions=extensions,
                        valid=valid,
                    )
                )
        return examples

    @staticmethod
    def all_test_catalogs() -> list[Catalog]:
        return [
            TestCases.case_1(),
            TestCases.case_2(),
            TestCases.case_3(),
            TestCases.case_4(),
            TestCases.case_5(),
            TestCases.case_7(),
            TestCases.case_8(),
        ]

    @staticmethod
    def case_1() -> Catalog:
        return Catalog.from_file(
            TestCases.get_path("data-files/catalogs/test-case-1/catalog.json")
        )

    @staticmethod
    def case_2() -> Catalog:
        return Catalog.from_file(
            TestCases.get_path("data-files/catalogs/test-case-2/catalog.json")
        )

    @staticmethod
    def case_3() -> Catalog:
        root_cat = Catalog(
            id="test3", description="test case 3 catalog", title="test case 3 title"
        )

        image_item = Item(
            id="imagery-item",
            geometry=ARBITRARY_GEOM,
            bbox=ARBITRARY_BBOX,
            datetime=datetime.now(timezone.utc),
            properties={},
        )

        image_item.add_asset(
            "ortho", Asset(href="some/geotiff.tiff", media_type=MediaType.GEOTIFF)
        )

        label_item = Item(
            id="label-items",
            geometry=ARBITRARY_GEOM,
            bbox=ARBITRARY_BBOX,
            datetime=datetime.now(timezone.utc),
            properties={},
        )

        root_cat.add_item(image_item)
        root_cat.add_item(label_item)

        return root_cat

    @staticmethod
    def case_4() -> Catalog:
        """Test case that is based on a local copy of the Tier 1 dataset from
        DrivenData's OpenCities AI Challenge.
        See: https://www.drivendata.org/competitions/60/building-segmentation-disaster\
-resilience
        """
        return Catalog.from_file(
            TestCases.get_path("data-files/catalogs/test-case-4/catalog.json")
        )

    @staticmethod
    def case_5() -> Catalog:
        """Based on a subset of https://cbers.stac.cloud/"""
        return Catalog.from_file(
            TestCases.get_path("data-files/catalogs/test-case-5/catalog.json")
        )

    @staticmethod
    def case_6() -> Catalog:
        """Based on a subset of CBERS, contains a root and 4 empty children"""
        return Catalog.from_file(
            TestCases.get_path("data-files/catalogs/cbers-partial/catalog.json")
        )

    @staticmethod
    def case_7() -> Catalog:
        """Test case 4 as STAC version 0.8.1"""
        return Catalog.from_file(
            TestCases.get_path("data-files/catalogs/label_catalog-v0.8.1/catalog.json")
        )

    @staticmethod
    def case_8() -> Collection:
        """Planet disaster data example catalog, 1.0.0-beta.2"""
        return Collection.from_file(
            TestCases.get_path(
                "data-files/catalogs/planet-example-v1.0.0-beta.2/collection.json"
            )
        )



"""Tests for pystac.tests.extensions.render"""

import json

import pytest

import pystac
import pystac.errors
from pystac.extensions.render import Render, RenderExtension
from tests.conftest import get_data_file


@pytest.fixture
def ext_collection_uri() -> str:
    return get_data_file("render/collection.json")


@pytest.fixture
def ext_collection(ext_collection_uri: str) -> pystac.Collection:
    return pystac.Collection.from_file(ext_collection_uri)


@pytest.fixture
def ext_item_uri() -> str:
    return get_data_file("render/item.json")


@pytest.fixture
def ext_item(ext_item_uri: str) -> pystac.Item:
    return pystac.Item.from_file(ext_item_uri)


@pytest.fixture
def render() -> Render:
    return Render.create(
        assets=["B4", "B3", "B2"],
        title="RGB",
        rescale=[[0, 1000], [0, 1000], [0, 1000]],
        nodata=-9999,
        colormap_name="viridis",
        colormap={
            "0": "#e5f5f9",
            "10": "#99d8c9",
            "255": "#2ca25f",
        },
        color_formula="gamma rg 1.3, sigmoidal rgb 22 0.1, saturation 1.5",
        resampling="bilinear",
        expression="(B08-B04)/(B08+B04)",
        minmax_zoom=[2, 18],
    )


@pytest.fixture
def thumbnail_render(ext_item: pystac.Item) -> Render:
    return RenderExtension.ext(ext_item).renders["thumbnail"]


def test_collection_stac_extensions(ext_collection: pystac.Collection) -> None:
    assert RenderExtension.has_extension(ext_collection)


def test_item_stac_extensions(ext_item: pystac.Item) -> None:
    assert RenderExtension.has_extension(ext_item)


def test_ext_raises_if_item_does_not_conform(item: pystac.Item) -> None:
    with pytest.raises(pystac.errors.ExtensionNotImplemented):
        RenderExtension.ext(item)


def test_ext_raises_if_collection_does_not_conform(
    collection: pystac.Collection,
) -> None:
    with pytest.raises(pystac.errors.ExtensionNotImplemented):
        RenderExtension.ext(collection)


def test_ext_raises_on_catalog(catalog: pystac.Catalog) -> None:
    with pytest.raises(
        pystac.errors.ExtensionTypeError,
        match="RenderExtension does not apply to type 'Catalog'",
    ):
        RenderExtension.ext(catalog)  # type: ignore


def test_item_to_from_dict(ext_item_uri: str, ext_item: pystac.Item) -> None:
    with open(ext_item_uri) as f:
        d = json.load(f)
    actual = ext_item.to_dict(include_self_link=False)
    assert actual == d


def test_collection_to_from_dict(
    ext_collection_uri: str, ext_collection: pystac.Collection
) -> None:
    with open(ext_collection_uri) as f:
        d = json.load(f)
    actual = ext_collection.to_dict(include_self_link=False)
    assert actual == d


def test_add_to_item(item: pystac.Item) -> None:
    assert not RenderExtension.has_extension(item)
    RenderExtension.add_to(item)

    assert RenderExtension.has_extension(item)


def test_add_to_collection(collection: pystac.Collection) -> None:
    assert not RenderExtension.has_extension(collection)
    RenderExtension.add_to(collection)

    assert RenderExtension.has_extension(collection)


def test_get_render_values(thumbnail_render: Render) -> None:
    assert thumbnail_render.title == "Thumbnail"
    assert thumbnail_render.assets == ["B04", "B03", "B02"]
    assert thumbnail_render.rescale == [[0, 150]]
    assert thumbnail_render.colormap_name == "rainbow"
    assert thumbnail_render.resampling == "bilinear"
    assert thumbnail_render.properties.get("bidx") == [1]
    assert thumbnail_render.properties.get("width") == 1024
    assert thumbnail_render.properties.get("height") == 1024
    assert thumbnail_render.properties.get("bands") == ["B4", "B3", "B2"]


def test_apply_renders_to_item(item: pystac.Item, render: Render) -> None:
    RenderExtension.add_to(item)

    RenderExtension.ext(item).apply({"render": render})
    assert item.ext.render.renders["render"].assets == ["B4", "B3", "B2"]
    assert item.ext.render.renders["render"].title == "RGB"
    assert item.ext.render.renders["render"].rescale == [
        [0, 1000],
        [0, 1000],
        [0, 1000],
    ]
    assert item.ext.render.renders["render"].nodata == -9999
    assert item.ext.render.renders["render"].colormap_name == "viridis"
    assert item.ext.render.renders["render"].colormap == {
        "0": "#e5f5f9",
        "10": "#99d8c9",
        "255": "#2ca25f",
    }
    assert (
        item.ext.render.renders["render"].color_formula
        == "gamma rg 1.3, sigmoidal rgb 22 0.1, saturation 1.5"
    )
    assert item.ext.render.renders["render"].resampling == "bilinear"
    assert item.ext.render.renders["render"].expression == "(B08-B04)/(B08+B04)"
    assert item.ext.render.renders["render"].minmax_zoom == [2, 18]

    assert item.ext.render.renders["render"] == render


def test_get_unset_properties(item: pystac.Item) -> None:
    RenderExtension.add_to(item)
    RenderExtension.ext(item).apply(
        {
            "render": Render.create(
                assets=["B4", "B3", "B2"],
            )
        }
    )

    assert item.ext.render.renders["render"].title is None
    assert item.ext.render.renders["render"].rescale is None
    assert item.ext.render.renders["render"].nodata is None
    assert item.ext.render.renders["render"].colormap_name is None
    assert item.ext.render.renders["render"].colormap is None
    assert item.ext.render.renders["render"].color_formula is None
    assert item.ext.render.renders["render"].resampling is None
    assert item.ext.render.renders["render"].expression is None
    assert item.ext.render.renders["render"].minmax_zoom is None


def test_equality_check_with_unexpected_type_raises_notimplemented_error() -> None:
    render = Render.create(
        assets=["B4", "B3", "B2"],
    )
    with pytest.raises(NotImplementedError):
        _ = render == 1


def test_item_repr(ext_item: pystac.Item) -> None:
    ext = RenderExtension.ext(ext_item)
    assert ext.__repr__() == f"<ItemRenderExtension Item id={ext_item.id}>"


def test_collection_repr(ext_collection: pystac.Collection) -> None:
    ext = RenderExtension.ext(ext_collection)
    assert (
        ext.__repr__()
        == f"<CollectionRenderExtension Collection id={ext_collection.id}>"
    )


def test_render_repr() -> None:
    render = Render.create(
        assets=["B4", "B3", "B2"],
        title="RGB",
        rescale=[[0, 1000], [0, 1000], [0, 1000]],
        nodata=-9999,
        colormap_name="viridis",
        colormap={
            "0": "#e5f5f9",
            "10": "#99d8c9",
            "255": "#2ca25f",
        },
        color_formula="gamma rg 1.3, sigmoidal rgb 22 0.1, saturation 1.5",
        resampling="bilinear",
        expression="(B08-B04)/(B08+B04)",
        minmax_zoom=[2, 18],
    )

    assert render.__repr__() == (
        "<Render assets=['B4', 'B3', 'B2'] "
        "title=RGB rescale=[[0, 1000], [0, 1000], [0, 1000]] "
        "nodata=-9999 colormap_name=viridis "
        "colormap={'0': '#e5f5f9', '10': '#99d8c9', '255': '#2ca25f'} "
        "color_formula=gamma rg 1.3, sigmoidal rgb 22 0.1, saturation 1.5 "
        "resampling=bilinear expression=(B08-B04)/(B08+B04) minmax_zoom=[2, 18]>"
    )


@pytest.mark.vcr
def test_item_validate(ext_item: pystac.Item) -> None:
    assert ext_item.validate()


@pytest.mark.vcr
def test_collection_validate(ext_collection: pystac.Collection) -> None:
    assert ext_collection.validate()



from pystac.utils import StringEnum


class MediaType(StringEnum):
    """A list of common media types that can be used in STAC Asset and Link metadata."""

    COG = "image/tiff; application=geotiff; profile=cloud-optimized"
    FLATGEOBUF = "application/vnd.flatgeobuf"  # https://github.com/flatgeobuf/flatgeobuf/discussions/112#discussioncomment-4606721  # noqa
    GEOJSON = "application/geo+json"
    GEOPACKAGE = "application/geopackage+sqlite3"
    GEOTIFF = "image/tiff; application=geotiff"
    HDF = "application/x-hdf"  # Hierarchical Data Format versions 4 and earlier.
    HDF5 = "application/x-hdf5"  # Hierarchical Data Format version 5
    HTML = "text/html"
    JPEG = "image/jpeg"
    JPEG2000 = "image/jp2"
    JSON = "application/json"
    PNG = "image/png"
    TEXT = "text/plain"
    TIFF = "image/tiff"
    KML = "application/vnd.google-earth.kml+xml"
    XML = "application/xml"
    PDF = "application/pdf"

    # https://github.com/Unidata/netcdf/issues/42#issuecomment-1007618822
    NETCDF = "application/netcdf"

    # Cloud Optimized Point Cloud
    COPC = "application/vnd.laszip+copc"

    # https://github.com/protomaps/PMTiles/blob/main/spec/v3/spec.md#1-abstract
    VND_PMTILES = "application/vnd.pmtiles"  # Protomaps PMTiles

    # https://www.iana.org/assignments/media-types/application/vnd.apache.parquet
    VND_APACHE_PARQUET = "application/vnd.apache.parquet"

    # https://humanbrainproject.github.io/openMINDS/v3/core/v4/data/contentType.html
    VND_ZARR = "application/vnd.zarr"

    ##############
    # DEPRECATED #
    ##############

    # https://github.com/opengeospatial/geoparquet/issues/115#issuecomment-1181549523
    # deprecated, use VND_APACHE_PARQUET instead
    PARQUET = "application/x-parquet"

    # https://github.com/openMetadataInitiative/openMINDS_core/blob/v4/instances/data/contentTypes/zarr.jsonld
    # deprecated, use VND_ZARR instead
    ZARR = "application/vnd+zarr"


#: Media types that can be resolved as STAC Objects
STAC_JSON = [None, MediaType.GEOJSON, MediaType.JSON]



import json

import pytest

import pystac
from pystac import ExtensionTypeError, Item
from pystac.errors import ExtensionNotImplemented, RequiredPropertyMissing
from pystac.extensions.eo import PREFIX, SNOW_COVER_PROP, Band, EOExtension
from pystac.extensions.projection import ProjectionExtension
from pystac.summaries import RangeSummary
from pystac.utils import get_opt
from tests.conftest import get_data_file
from tests.utils import TestCases, assert_to_from_dict


def test_band_create() -> None:
    band = Band.create(
        name="B01",
        common_name="red",
        description=Band.band_description("red"),
        center_wavelength=0.65,
        full_width_half_max=0.1,
        solar_illumination=42.0,
    )

    assert band.name == "B01"
    assert band.common_name == "red"
    assert band.description, "Common name: red == Range: 0.6 to 0.7"
    assert band.center_wavelength == 0.65
    assert band.full_width_half_max == 0.1
    assert band.solar_illumination == 42.0
    assert band.__repr__() == "<Band name=B01>"


def test_band_description_unknown_band() -> None:
    desc = Band.band_description("rainbow")
    assert desc is None


LANDSAT_EXAMPLE_URI = TestCases.get_path("data-files/eo/eo-landsat-example.json")
BANDS_IN_ITEM_URI = TestCases.get_path(
    "data-files/eo/sample-bands-in-item-properties.json"
)
EO_COLLECTION_URI = TestCases.get_path("data-files/eo/eo-collection.json")
S2_ITEM_URI = TestCases.get_path("data-files/eo/eo-sentinel2-item.json")
PLAIN_ITEM = TestCases.get_path("data-files/item/sample-item.json")


def test_to_from_dict() -> None:
    with open(LANDSAT_EXAMPLE_URI) as f:
        item_dict = json.load(f)
    assert_to_from_dict(Item, item_dict)


def test_from_file_to_dict(ext_item_uri: str, ext_item: pystac.Item) -> None:
    with open(ext_item_uri) as f:
        d = json.load(f)
    actual = ext_item.to_dict(include_self_link=False)
    assert actual == d


def test_add_to() -> None:
    item = Item.from_file(PLAIN_ITEM)
    assert EOExtension.get_schema_uri() not in item.stac_extensions

    # Check that the URI gets added to stac_extensions
    EOExtension.add_to(item)
    assert EOExtension.get_schema_uri() in item.stac_extensions

    # Check that the URI only gets added once, regardless of how many times add_to
    # is called.
    EOExtension.add_to(item)
    EOExtension.add_to(item)

    eo_uris = [
        uri for uri in item.stac_extensions if uri == EOExtension.get_schema_uri()
    ]
    assert len(eo_uris) == 1


@pytest.mark.vcr()
def test_validate_eo() -> None:
    item = pystac.Item.from_file(LANDSAT_EXAMPLE_URI)
    item2 = pystac.Item.from_file(BANDS_IN_ITEM_URI)
    item.validate()
    item2.validate()


@pytest.mark.vcr()
def test_bands() -> None:
    item = pystac.Item.from_file(BANDS_IN_ITEM_URI)

    # Get
    assert "eo:bands" in item.properties
    bands = EOExtension.ext(item).bands
    assert bands is not None
    assert list(map(lambda x: x.name, bands)) == ["band1", "band2", "band3", "band4"]

    # Set
    new_bands = [
        Band.create(name="red", description=Band.band_description("red")),
        Band.create(name="green", description=Band.band_description("green")),
        Band.create(name="blue", description=Band.band_description("blue")),
    ]

    EOExtension.ext(item).bands = new_bands
    assert (
        "Common name: red, Range: 0.6 to 0.7"
        == item.properties["eo:bands"][0]["description"]
    )
    assert len(EOExtension.ext(item).bands or []) == 3
    item.validate()


def test_asset_bands_s2() -> None:
    item = pystac.Item.from_file(S2_ITEM_URI)
    mtd_asset = item.get_assets()["mtd"]
    assert EOExtension.ext(mtd_asset).bands is None


@pytest.mark.vcr()
def test_asset_bands() -> None:
    item = pystac.Item.from_file(LANDSAT_EXAMPLE_URI)

    # Get

    b1_asset = item.assets["B1"]
    asset_bands = EOExtension.ext(b1_asset).bands
    assert asset_bands is not None
    assert len(asset_bands) == 1
    assert asset_bands[0].name == "B1"
    assert asset_bands[0].solar_illumination == 2000

    index_asset = item.assets["index"]
    asset_bands = EOExtension.ext(index_asset).bands
    assert asset_bands is None

    # No asset specified
    item_bands = EOExtension.ext(item).bands
    assert item_bands is not None

    # Set
    b2_asset = item.assets["B2"]
    assert get_opt(EOExtension.ext(b2_asset).bands)[0].name == "B2"
    EOExtension.ext(b2_asset).bands = EOExtension.ext(b1_asset).bands

    new_b2_asset_bands = EOExtension.ext(item.assets["B2"]).bands

    assert get_opt(new_b2_asset_bands)[0].name == "B1"

    item.validate()

    # Check adding a new asset
    new_bands = [
        Band.create(
            name="red",
            description=Band.band_description("red"),
            solar_illumination=1900,
        ),
        Band.create(
            name="green",
            description=Band.band_description("green"),
            solar_illumination=1950,
        ),
        Band.create(
            name="blue",
            description=Band.band_description("blue"),
            solar_illumination=2000,
        ),
    ]
    asset = pystac.Asset(href="some/path.tif", media_type=pystac.MediaType.GEOTIFF)
    EOExtension.ext(asset).bands = new_bands
    item.add_asset("test", asset)

    assert len(item.assets["test"].extra_fields["eo:bands"]) == 3


@pytest.mark.vcr()
def test_cloud_cover() -> None:
    item = pystac.Item.from_file(LANDSAT_EXAMPLE_URI)

    # Get
    assert "eo:cloud_cover" in item.properties
    cloud_cover = EOExtension.ext(item).cloud_cover
    assert cloud_cover == 78

    # Set
    EOExtension.ext(item).cloud_cover = 50
    assert item.properties["eo:cloud_cover"] == 50

    # Get from Asset
    b2_asset = item.assets["B2"]
    assert EOExtension.ext(b2_asset).cloud_cover == EOExtension.ext(item).cloud_cover

    b3_asset = item.assets["B3"]
    assert EOExtension.ext(b3_asset).cloud_cover == 20

    # Set on Asset
    EOExtension.ext(b2_asset).cloud_cover = 10
    assert EOExtension.ext(b2_asset).cloud_cover == 10

    item.validate()


def test_summaries() -> None:
    col = pystac.Collection.from_file(EO_COLLECTION_URI)
    eo_summaries = EOExtension.summaries(col)

    # Get

    cloud_cover_summaries = eo_summaries.cloud_cover
    assert cloud_cover_summaries is not None
    assert cloud_cover_summaries.minimum == 0.0
    assert cloud_cover_summaries.maximum == 80.0

    snow_cover_summaries = eo_summaries.snow_cover
    assert snow_cover_summaries is not None
    assert snow_cover_summaries.minimum == 0.0
    assert snow_cover_summaries.maximum == 80.0

    bands = eo_summaries.bands
    assert bands is not None
    assert len(bands) == 11

    # Set

    eo_summaries.cloud_cover = RangeSummary(1.0, 2.0)
    eo_summaries.snow_cover = RangeSummary(4.0, 23)
    eo_summaries.bands = [Band.create(name="test")]

    col_dict = col.to_dict()
    assert len(col_dict["summaries"]["eo:bands"]) == 1
    assert col_dict["summaries"]["eo:cloud_cover"]["minimum"] == 1.0
    assert col_dict["summaries"]["eo:snow_cover"]["minimum"] == 4.0


def test_summaries_adds_uri() -> None:
    col = pystac.Collection.from_file(EO_COLLECTION_URI)
    col.stac_extensions = []
    with pytest.raises(
        pystac.ExtensionNotImplemented, match="Extension 'eo' is not implemented"
    ):
        EOExtension.summaries(col, add_if_missing=False)

    EOExtension.summaries(col, add_if_missing=True)

    assert EOExtension.get_schema_uri() in col.stac_extensions

    EOExtension.remove_from(col)
    assert EOExtension.get_schema_uri() not in col.stac_extensions


def test_read_pre_09_fields_into_common_metadata() -> None:
    eo_item = pystac.Item.from_file(
        TestCases.get_path(
            "data-files/examples/0.8.1/item-spec/examples/landsat8-sample.json"
        )
    )

    assert eo_item.common_metadata.platform == "landsat-8"
    assert eo_item.common_metadata.instruments == ["oli_tirs"]


def test_reads_asset_bands_in_pre_1_0_version() -> None:
    item = pystac.Item.from_file(
        TestCases.get_path(
            "data-files/examples/0.9.0/item-spec/examples/landsat8-sample.json"
        )
    )

    bands = EOExtension.ext(item.assets["B9"]).bands

    assert len(bands or []) == 1
    assert get_opt(bands)[0].common_name == "cirrus"


def test_reads_gsd_in_pre_1_0_version() -> None:
    eo_item = pystac.Item.from_file(
        TestCases.get_path(
            "data-files/examples/0.9.0/item-spec/examples/landsat8-sample.json"
        )
    )

    assert eo_item.common_metadata.gsd == 30.0


def test_item_apply() -> None:
    item = pystac.Item.from_file(LANDSAT_EXAMPLE_URI)
    eo_ext = EOExtension.ext(item)
    test_band = Band.create(name="test")

    assert eo_ext.cloud_cover == 78
    assert test_band not in (eo_ext.bands or [])

    eo_ext.apply(bands=[test_band], cloud_cover=15)
    assert eo_ext.bands is not None

    assert test_band.to_dict() == eo_ext.bands[0].to_dict()
    assert eo_ext.cloud_cover == 15


def test_extend_invalid_object() -> None:
    link = pystac.Link("child", "https://some-domain.com/some/path/to.json")

    with pytest.raises(pystac.ExtensionTypeError):
        EOExtension.ext(link)  # type: ignore


def test_extension_not_implemented() -> None:
    # Should raise exception if Item does not include extension URI
    item = pystac.Item.from_file(PLAIN_ITEM)

    with pytest.raises(pystac.ExtensionNotImplemented):
        _ = EOExtension.ext(item)

    # Should raise exception if owning Item does not include extension URI
    asset = item.assets["thumbnail"]

    with pytest.raises(pystac.ExtensionNotImplemented):
        _ = EOExtension.ext(asset)

    # Should succeed if Asset has no owner
    ownerless_asset = pystac.Asset.from_dict(asset.to_dict())
    _ = EOExtension.ext(ownerless_asset)


def test_item_ext_add_to() -> None:
    item = pystac.Item.from_file(PLAIN_ITEM)
    assert EOExtension.get_schema_uri() not in item.stac_extensions

    _ = EOExtension.ext(item, add_if_missing=True)

    assert EOExtension.get_schema_uri() in item.stac_extensions


def test_asset_ext_add_to() -> None:
    item = pystac.Item.from_file(PLAIN_ITEM)
    assert EOExtension.get_schema_uri() not in item.stac_extensions
    asset = item.assets["thumbnail"]

    _ = EOExtension.ext(asset, add_if_missing=True)

    assert EOExtension.get_schema_uri() in item.stac_extensions


def test_asset_ext_add_to_ownerless_asset() -> None:
    item = pystac.Item.from_file(PLAIN_ITEM)
    asset_dict = item.assets["thumbnail"].to_dict()
    asset = pystac.Asset.from_dict(asset_dict)

    with pytest.raises(pystac.STACError):
        _ = EOExtension.ext(asset, add_if_missing=True)


def test_should_raise_exception_when_passing_invalid_extension_object() -> None:
    with pytest.raises(
        ExtensionTypeError, match=r"^EOExtension does not apply to type 'object'$"
    ):
        # calling it wrong purposely -------v
        EOExtension.ext(object())  # type: ignore


def test_migration() -> None:
    item_0_8_path = TestCases.get_path(
        "data-files/examples/0.8.1/item-spec/examples/sentinel2-sample.json"
    )
    with open(item_0_8_path) as src:
        item_dict = json.load(src)

    assert "eo:epsg" in item_dict["properties"]

    item = Item.from_file(item_0_8_path)

    assert "eo:epsg" not in item.properties
    assert "proj:code" in item.properties
    assert ProjectionExtension.get_schema_uri() in item.stac_extensions
    assert item.ext.proj.epsg == item_dict["properties"]["eo:epsg"]


@pytest.fixture
def ext_item_uri() -> str:
    return get_data_file("eo/eo-landsat-example.json")


@pytest.fixture
def ext_item(ext_item_uri: str) -> pystac.Item:
    return pystac.Item.from_file(ext_item_uri)


@pytest.mark.parametrize("field", ["cloud_cover", "snow_cover"])
def test_get_field(ext_item: pystac.Item, field: str) -> None:
    prop = ext_item.properties[f"{PREFIX}{field}"]
    attr = getattr(EOExtension.ext(ext_item), field)

    assert attr is not None
    assert attr == prop


@pytest.mark.vcr()
@pytest.mark.parametrize(
    "field,value",
    [
        ("cloud_cover", 7.8),
        ("snow_cover", 99),
    ],
)
def test_set_field(ext_item: pystac.Item, field: str, value) -> None:  # type: ignore
    original = ext_item.properties[f"{PREFIX}{field}"]
    setattr(EOExtension.ext(ext_item), field, value)
    new = ext_item.properties[f"{PREFIX}{field}"]

    assert new != original
    assert new == value
    assert ext_item.validate()


def test_snow_cover_set_to_none_pops_from_dict(ext_item: pystac.Item) -> None:
    assert SNOW_COVER_PROP in ext_item.properties

    EOExtension.ext(ext_item).snow_cover = None
    assert SNOW_COVER_PROP not in ext_item.properties


def test_snow_cover_raises_informative_error(ext_item: pystac.Item) -> None:
    with pytest.raises(ValueError, match="must be number"):
        EOExtension.ext(ext_item).snow_cover = True

    with pytest.raises(ValueError, match="must be between 0 and 100"):
        EOExtension.ext(ext_item).snow_cover = 100.5


def test_older_extension_version(ext_item: Item) -> None:
    old = "https://stac-extensions.github.io/eo/v1.0.0/schema.json"
    new = "https://stac-extensions.github.io/eo/v1.1.0/schema.json"

    stac_extensions = set(ext_item.stac_extensions)
    stac_extensions.remove(new)
    stac_extensions.add(old)
    item_as_dict = ext_item.to_dict(include_self_link=False, transform_hrefs=False)
    item_as_dict["stac_extensions"] = list(stac_extensions)
    item = Item.from_dict(item_as_dict, migrate=False)
    assert EOExtension.has_extension(item)
    assert old in item.stac_extensions

    migrated_item = pystac.Item.from_dict(item_as_dict, migrate=True)
    assert EOExtension.has_extension(migrated_item)
    assert new in migrated_item.stac_extensions


@pytest.mark.parametrize(
    "filter,count",
    [
        (dict(), 11),
        ({"name": "B4"}, 1),
        ({"common_name": "blue"}, 1),
        ({"name": "B4", "common_name": "red"}, 1),
        ({"name": "B4", "common_name": "green"}, 0),
    ],
)
def test_get_assets(ext_item: pystac.Item, filter: dict[str, str], count: int) -> None:
    assets = EOExtension.ext(ext_item).get_assets(**filter)  # type:ignore
    assert len(assets) == count


def test_get_assets_works_even_if_band_info_is_incomplete(
    ext_item: pystac.Item,
) -> None:
    name = ext_item.assets["B4"].extra_fields["eo:bands"][0].pop("name")
    common_name = ext_item.assets["B2"].extra_fields["eo:bands"][0].pop("common_name")

    eo_ext = EOExtension.ext(ext_item)

    assets = eo_ext.get_assets(name=name)  # type:ignore
    assert len(assets) == 0

    assets = eo_ext.get_assets(common_name=common_name)  # type:ignore
    assert len(assets) == 0


def test_exception_should_include_hint_if_obj_is_collection(
    collection: pystac.Collection,
) -> None:
    with pytest.raises(
        ExtensionTypeError,
        match="Hint: Did you mean to use `EOExtension.summaries` instead?",
    ):
        EOExtension.ext(collection)  # type:ignore


def test_ext_syntax(ext_item: pystac.Item) -> None:
    assert ext_item.ext.eo.cloud_cover == 78
    assert (bands := ext_item.assets["B1"].ext.eo.bands)
    assert bands[0].name == "B1"


def test_ext_syntax_remove(ext_item: pystac.Item) -> None:
    ext_item.ext.remove("eo")
    assert ext_item.ext.has("eo") is False
    with pytest.raises(ExtensionNotImplemented):
        ext_item.ext.eo


def test_ext_syntax_add(item: pystac.Item) -> None:
    item.ext.add("eo")
    assert item.ext.has("eo") is True
    assert isinstance(item.ext.eo, EOExtension)


def test_required_property_missing(ext_item: pystac.Item) -> None:
    # https://github.com/stac-utils/pystac/issues/1402
    d = ext_item.to_dict(include_self_link=False, transform_hrefs=False)
    del d["assets"]["B1"]["eo:bands"][0]["name"]
    item = pystac.Item.from_dict(d)
    bands = item.assets["B1"].ext.eo.bands
    assert bands is not None
    with pytest.raises(RequiredPropertyMissing):
        bands[0].name


def test_unnecessary_migrations_not_performed(ext_item: Item) -> None:
    item_as_dict = ext_item.to_dict(include_self_link=False, transform_hrefs=False)
    item_as_dict["stac_version"] = "1.0.0"
    item_as_dict["properties"]["eo:bands"] = [{"name": "B1", "common_name": "coastal"}]

    item = Item.from_dict(item_as_dict)

    migrated_item = pystac.Item.from_dict(item_as_dict, migrate=True)

    assert item.properties == migrated_item.properties
    assert len(item.assets) == len(migrated_item.assets)
    for key, value in item.assets.items():
        assert value.to_dict() == migrated_item.assets[key].to_dict()



"""Implements the Storage Extension.

https://github.com/stac-extensions/storage
"""

from __future__ import annotations

from collections.abc import Iterable
from typing import (
    Any,
    Generic,
    Literal,
    TypeVar,
    cast,
)

import pystac
from pystac.extensions.base import (
    ExtensionManagementMixin,
    PropertiesExtension,
    SummariesExtension,
)
from pystac.extensions.hooks import ExtensionHooks
from pystac.utils import StringEnum

#: Generalized version of :class:`~pystac.Item`, :class:`~pystac.Asset` or
#: :class:`~pystac.ItemAssetDefinition`
T = TypeVar("T", pystac.Item, pystac.Asset, pystac.ItemAssetDefinition)

SCHEMA_URI: str = "https://stac-extensions.github.io/storage/v1.0.0/schema.json"
PREFIX: str = "storage:"

# Field names
PLATFORM_PROP: str = PREFIX + "platform"
REGION_PROP: str = PREFIX + "region"
REQUESTER_PAYS_PROP: str = PREFIX + "requester_pays"
TIER_PROP: str = PREFIX + "tier"


class CloudPlatform(StringEnum):
    ALIBABA = "ALIBABA"
    AWS = "AWS"
    AZURE = "AZURE"
    GCP = "GCP"
    IBM = "IBM"
    ORACLE = "ORACLE"
    OTHER = "OTHER"


class StorageExtension(
    Generic[T],
    PropertiesExtension,
    ExtensionManagementMixin[pystac.Item | pystac.Collection],
):
    """An abstract class that can be used to extend the properties of an
    :class:`~pystac.Item` or :class:`~pystac.Asset` with properties from the
    :stac-ext:`Storage Extension <storage>`. This class is generic over the type of
    STAC Object to be extended (e.g. :class:`~pystac.Item`,
    :class:`~pystac.Asset`).

    To create a concrete instance of :class:`StorageExtension`, use the
    :meth:`StorageExtension.ext` method. For example:

    .. code-block:: python

       >>> item: pystac.Item = ...
       >>> storage_ext = StorageExtension.ext(item)
    """

    name: Literal["storage"] = "storage"

    def apply(
        self,
        platform: CloudPlatform | None = None,
        region: str | None = None,
        requester_pays: bool | None = None,
        tier: str | None = None,
    ) -> None:
        """Applies Storage Extension properties to the extended :class:`~pystac.Item` or
        :class:`~pystac.Asset`.

        Args:
            platform (str, CloudPlatform) : The cloud provider where data is stored.
            region (str) : The region where the data is stored. Relevant to speed of
                access and inter-region egress costs (as defined by PaaS provider).
            requester_pays (bool) : Is the data requester pays or is it data
                manager/cloud provider pays.
            tier (str) : The title for the tier type (as defined by PaaS provider).
        """
        self.platform = platform
        self.region = region
        self.requester_pays = requester_pays
        self.tier = tier

    @property
    def platform(self) -> CloudPlatform | None:
        """Get or sets the cloud provider where data is stored.

        Returns:
            str or None
        """
        return self._get_property(PLATFORM_PROP, CloudPlatform)

    @platform.setter
    def platform(self, v: CloudPlatform | None) -> None:
        self._set_property(PLATFORM_PROP, v)

    @property
    def region(self) -> str | None:
        """Gets or sets the region where the data is stored. Relevant to speed of
        access and inter-region egress costs (as defined by PaaS provider)."""
        return self._get_property(REGION_PROP, str)

    @region.setter
    def region(self, v: str | None) -> None:
        self._set_property(REGION_PROP, v)

    @property
    def requester_pays(self) -> bool | None:
        # This value "defaults to false", according to the extension spec.
        return self._get_property(REQUESTER_PAYS_PROP, bool)

    @requester_pays.setter
    def requester_pays(self, v: bool | None) -> None:
        self._set_property(REQUESTER_PAYS_PROP, v)

    @property
    def tier(self) -> str | None:
        return self._get_property(TIER_PROP, str)

    @tier.setter
    def tier(self, v: str | None) -> None:
        self._set_property(TIER_PROP, v)

    @classmethod
    def get_schema_uri(cls) -> str:
        return SCHEMA_URI

    @classmethod
    def ext(cls, obj: T, add_if_missing: bool = False) -> StorageExtension[T]:
        """Extends the given STAC Object with properties from the :stac-ext:`Storage
        Extension <storage>`.

        This extension can be applied to instances of :class:`~pystac.Item` or
        :class:`~pystac.Asset`.

        Raises:

            pystac.ExtensionTypeError : If an invalid object type is passed.
        """
        if isinstance(obj, pystac.Item):
            cls.ensure_has_extension(obj, add_if_missing)
            return cast(StorageExtension[T], ItemStorageExtension(obj))
        elif isinstance(obj, pystac.Asset):
            cls.ensure_owner_has_extension(obj, add_if_missing)
            return cast(StorageExtension[T], AssetStorageExtension(obj))
        elif isinstance(obj, pystac.ItemAssetDefinition):
            cls.ensure_owner_has_extension(obj, add_if_missing)
            return cast(StorageExtension[T], ItemAssetsStorageExtension(obj))
        else:
            raise pystac.ExtensionTypeError(cls._ext_error_message(obj))

    @classmethod
    def summaries(
        cls, obj: pystac.Collection, add_if_missing: bool = False
    ) -> SummariesStorageExtension:
        """Returns the extended summaries object for the given collection."""
        cls.ensure_has_extension(obj, add_if_missing)
        return SummariesStorageExtension(obj)


class ItemStorageExtension(StorageExtension[pystac.Item]):
    """A concrete implementation of :class:`StorageExtension` on an
    :class:`~pystac.Item` that extends the properties of the Item to include
    properties defined in the :stac-ext:`Storage Extension <storage>`.

    This class should generally not be instantiated directly. Instead, call
    :meth:`StorageExtension.ext` on an :class:`~pystac.Item` to extend it.
    """

    item: pystac.Item
    """The :class:`~pystac.Item` being extended."""

    properties: dict[str, Any]
    """The :class:`~pystac.Item` properties, including extension properties."""

    def __init__(self, item: pystac.Item):
        self.item = item
        self.properties = item.properties

    def __repr__(self) -> str:
        return f"<ItemStorageExtension Item id={self.item.id}>"


class AssetStorageExtension(StorageExtension[pystac.Asset]):
    """A concrete implementation of :class:`StorageExtension` on an
    :class:`~pystac.Asset` that extends the Asset fields to include properties defined
    in the :stac-ext:`Storage Extension <storage>`.

    This class should generally not be instantiated directly. Instead, call
    :meth:`StorageExtension.ext` on an :class:`~pystac.Asset` to extend it.
    """

    asset_href: str
    """The ``href`` value of the :class:`~pystac.Asset` being extended."""

    properties: dict[str, Any]
    """The :class:`~pystac.Asset` fields, including extension properties."""

    additional_read_properties: Iterable[dict[str, Any]] | None = None
    """If present, this will be a list containing 1 dictionary representing the
    properties of the owning :class:`~pystac.Item`."""

    def __init__(self, asset: pystac.Asset):
        self.asset_href = asset.href
        self.properties = asset.extra_fields
        if asset.owner and isinstance(asset.owner, pystac.Item):
            self.additional_read_properties = [asset.owner.properties]

    def __repr__(self) -> str:
        return f"<AssetStorageExtension Asset href={self.asset_href}>"


class ItemAssetsStorageExtension(StorageExtension[pystac.ItemAssetDefinition]):
    properties: dict[str, Any]
    asset_defn: pystac.ItemAssetDefinition

    def __init__(self, item_asset: pystac.ItemAssetDefinition):
        self.asset_defn = item_asset
        self.properties = item_asset.properties


class SummariesStorageExtension(SummariesExtension):
    """A concrete implementation of :class:`~pystac.extensions.base.SummariesExtension`
    that extends the ``summaries`` field of a :class:`~pystac.Collection` to include
    properties defined in the :stac-ext:`Storage Extension <storage>`.
    """

    @property
    def platform(self) -> list[CloudPlatform] | None:
        """Get or sets the summary of :attr:`StorageExtension.platform` values
        for this Collection.
        """
        return self.summaries.get_list(PLATFORM_PROP)

    @platform.setter
    def platform(self, v: list[CloudPlatform] | None) -> None:
        self._set_summary(PLATFORM_PROP, v)

    @property
    def region(self) -> list[str] | None:
        """Get or sets the summary of :attr:`StorageExtension.region` values
        for this Collection.
        """
        return self.summaries.get_list(REGION_PROP)

    @region.setter
    def region(self, v: list[str] | None) -> None:
        self._set_summary(REGION_PROP, v)

    @property
    def requester_pays(self) -> list[bool] | None:
        """Get or sets the summary of :attr:`StorageExtension.requester_pays` values
        for this Collection.
        """
        return self.summaries.get_list(REQUESTER_PAYS_PROP)

    @requester_pays.setter
    def requester_pays(self, v: list[bool] | None) -> None:
        self._set_summary(REQUESTER_PAYS_PROP, v)

    @property
    def tier(self) -> list[str] | None:
        """Get or sets the summary of :attr:`StorageExtension.tier` values
        for this Collection.
        """
        return self.summaries.get_list(TIER_PROP)

    @tier.setter
    def tier(self, v: list[str] | None) -> None:
        self._set_summary(TIER_PROP, v)


class StorageExtensionHooks(ExtensionHooks):
    schema_uri: str = SCHEMA_URI
    prev_extension_ids: set[str] = set()
    stac_object_types = {
        pystac.STACObjectType.COLLECTION,
        pystac.STACObjectType.ITEM,
    }


STORAGE_EXTENSION_HOOKS: ExtensionHooks = StorageExtensionHooks()



import pytest

from pystac import Collection
from pystac.errors import DeprecatedWarning
from pystac.extensions.item_assets import AssetDefinition, ItemAssetsExtension
from pystac.item_assets import ItemAssetDefinition
from tests.utils import TestCases

CLASSIFICATION_COLLECTION_RASTER_URI = TestCases.get_path(
    "data-files/classification/collection-item-assets-raster-bands.json"
)


@pytest.fixture
def landsat8_collection() -> Collection:
    return Collection.from_file(
        TestCases.get_path("data-files/item-assets/example-landsat8.json")
    )


def test_example(landsat8_collection: Collection) -> None:
    assert len(landsat8_collection.item_assets) == 13

    assert landsat8_collection.item_assets["B1"] == ItemAssetDefinition(
        {
            "type": "image/tiff; application=geotiff",
            "eo:bands": [
                {
                    "name": "B1",
                    "common_name": "coastal",
                    "center_wavelength": 0.44,
                    "full_width_half_max": 0.02,
                }
            ],
            "title": "Coastal Band (B1)",
            "description": "Coastal Band Top Of the Atmosphere",
        }
    )


def test_set_using_dict(landsat8_collection: Collection) -> None:
    assert len(landsat8_collection.item_assets) == 13

    landsat8_collection.item_assets["Bx"] = {
        "type": "image/tiff; application=geotiff",
        "eo:bands": [
            {
                "name": "B1",
                "common_name": "coastal",
                "center_wavelength": 0.44,
                "full_width_half_max": 0.02,
            }
        ],
        "title": "Coastal Band (B1)",
        "description": "Coastal Band Top Of the Atmosphere",
    }  # type:ignore

    assert (
        landsat8_collection.item_assets["B1"] == landsat8_collection.item_assets["Bx"]
    )


class TestAssetDefinition:
    def test_eq(self, landsat8_collection: Collection) -> None:
        assert landsat8_collection.item_assets["B1"] != {"title": "Coastal Band (B1)"}

    def test_create(self) -> None:
        title = "Coastal Band (B1)"
        description = "Coastal Band Top Of the Atmosphere"
        media_type = "image/tiff; application=geotiff"
        roles = ["data"]
        asset_defn = ItemAssetDefinition.create(
            title=title, description=description, media_type=media_type, roles=roles
        )
        assert (
            asset_defn.title,
            asset_defn.description,
            asset_defn.media_type,
            asset_defn.roles,
        ) == (title, description, media_type, roles)

    def test_title(self) -> None:
        asset_defn = ItemAssetDefinition({})
        title = "Very Important Asset"

        asset_defn.title = title

        assert asset_defn.title == asset_defn.to_dict()["title"] == title

    def test_description(self) -> None:
        asset_defn = ItemAssetDefinition({})
        description = "What an incredibly important asset this is!"

        asset_defn.description = description

        assert (
            asset_defn.description == asset_defn.to_dict()["description"] == description
        )

    def test_media_type(self) -> None:
        asset_defn = ItemAssetDefinition({})
        media_type = "application/json"

        asset_defn.media_type = media_type

        assert asset_defn.media_type == asset_defn.to_dict()["type"] == media_type

    def test_roles(self) -> None:
        asset_defn = ItemAssetDefinition({})
        roles = ["thumbnail"]

        asset_defn.roles = roles

        assert asset_defn.roles == asset_defn.to_dict()["roles"] == roles

    def test_set_owner(self, landsat8_collection: Collection) -> None:
        asset_definition = ItemAssetDefinition(
            {
                "type": "image/tiff; application=geotiff",
                "eo:bands": [
                    {
                        "name": "B1",
                        "common_name": "coastal",
                        "center_wavelength": 0.44,
                        "full_width_half_max": 0.02,
                    }
                ],
                "title": "Coastal Band (B1)",
                "description": "Coastal Band Top Of the Atmosphere",
            }
        )
        asset_definition.set_owner(landsat8_collection)
        assert asset_definition.owner == landsat8_collection


def test_extra_fields(collection: Collection) -> None:
    asset_definition = ItemAssetDefinition.create(
        title=None,
        description=None,
        media_type=None,
        roles=None,
        extra_fields={"raster:bands": [{"nodata": 42}]},
    )

    collection.item_assets = {"data": asset_definition}
    assert collection.item_assets["data"].owner == collection

    collection_as_dict = collection.to_dict()
    assert collection_as_dict["item_assets"]["data"]["raster:bands"] == [{"nodata": 42}]
    asset = asset_definition.create_asset("asset.tif")
    assert asset.extra_fields["raster:bands"] == [{"nodata": 42}]

    collection.item_assets["data"].ext.add("raster")
    assert (bands := collection.item_assets["data"].ext.raster.bands)
    assert bands[0].nodata == 42

    assert collection.item_assets["data"].ext.has("raster")
    assert collection.ext.has("raster")


def test_set_item_asset(collection: Collection) -> None:
    asset_definition = ItemAssetDefinition.create(
        title=None,
        description=None,
        media_type=None,
        roles=None,
        extra_fields={"raster:bands": [{"nodata": 42}]},
    )

    collection.item_assets["data"] = asset_definition
    assert collection.item_assets["data"].owner == collection


def test_item_assets_extension_is_deprecated() -> None:
    collection = Collection.from_file(CLASSIFICATION_COLLECTION_RASTER_URI)

    assert ItemAssetsExtension.get_schema_uri() not in collection.stac_extensions

    with pytest.warns(DeprecatedWarning, match="top-level property of"):
        item_asset_ext = ItemAssetsExtension.ext(collection, add_if_missing=True)
        item_asset = item_asset_ext.item_assets["cloud-mask-raster"]

    assert collection.id in repr(item_asset_ext)

    assert item_asset.ext.has("eo")

    with pytest.warns(DeprecatedWarning, match="top-level property of"):
        assert collection.ext.item_assets["cloud-mask-raster"].ext.has("eo")

    assert ItemAssetsExtension.get_schema_uri() in collection.stac_extensions

    with pytest.warns(DeprecationWarning):
        asset_definition = AssetDefinition(
            {"title": "Thumbnail image", "type": "image/jpeg"}
        )
    item_asset_ext.item_assets["thumbnail"] = asset_definition


def test_item_assets_extension_asset_definition_is_deprecated() -> None:
    with pytest.warns(
        DeprecationWarning, match="Please use ``pystac.ItemAssetDefinition``"
    ):
        asset_definition = AssetDefinition(
            {
                "type": "image/tiff; application=geotiff",
                "eo:bands": [
                    {
                        "name": "B1",
                        "common_name": "coastal",
                        "center_wavelength": 0.44,
                        "full_width_half_max": 0.02,
                    }
                ],
                "title": "Coastal Band (B1)",
                "description": "Coastal Band Top Of the Atmosphere",
            }
        )

    assert asset_definition.title == "Coastal Band (B1)"
    assert asset_definition.ext.eo.bands
    assert asset_definition.ext.eo.bands[0].name == "B1"
    assert asset_definition.owner is None



import json
from typing import Any

import pytest

import pystac
from pystac.asset import Asset
from pystac.errors import ExtensionTypeError, RequiredPropertyMissing, STACError
from pystac.extensions.pointcloud import (
    AssetPointcloudExtension,
    PhenomenologyType,
    PointcloudExtension,
    Schema,
    SchemaType,
    Statistic,
)
from pystac.summaries import RangeSummary
from tests.utils import TestCases, assert_to_from_dict

# from copy import deepcopy


@pytest.fixture
def example_uri() -> str:
    return TestCases.get_path("data-files/pointcloud/example-laz.json")


@pytest.fixture
def pc_item(example_uri: str) -> pystac.Item:
    return pystac.Item.from_file(example_uri)


@pytest.fixture
def pc_no_stats_item() -> pystac.Item:
    return pystac.Item.from_file(
        TestCases.get_path("data-files/pointcloud/example-laz-no-statistics.json")
    )


@pytest.fixture
def plain_item() -> pystac.Item:
    return pystac.Item.from_file(TestCases.get_path("data-files/item/sample-item.json"))


@pytest.fixture
def collection() -> pystac.Collection:
    return pystac.Collection.from_file(
        TestCases.get_path("data-files/collections/multi-extent.json")
    )


def test_to_from_dict(example_uri: str) -> None:
    with open(example_uri) as f:
        d = json.load(f)
    assert_to_from_dict(pystac.Item, d)


def test_apply() -> None:
    item = next(TestCases.case_2().get_items(recursive=True))

    assert not PointcloudExtension.has_extension(item)

    PointcloudExtension.add_to(item)
    PointcloudExtension.ext(item).apply(
        1000,
        PhenomenologyType.LIDAR,
        "laszip",
        [Schema({"name": "X", "size": 8, "type": "floating"})],
    )
    assert PointcloudExtension.has_extension(item)


@pytest.mark.vcr()
def test_validate_pointcloud(pc_item: pystac.Item) -> None:
    pc_item.validate()


@pytest.mark.vcr()
def test_count(pc_item: pystac.Item) -> None:
    # Get
    assert "pc:count" in pc_item.properties
    pc_count = PointcloudExtension.ext(pc_item).count
    assert pc_count == pc_item.properties["pc:count"]

    # Set
    PointcloudExtension.ext(pc_item).count = pc_count + 100
    assert pc_count + 100 == pc_item.properties["pc:count"]

    # Validate
    pc_item.validate()

    # Cannot test validation errors until the pointcloud schema.json syntax is fixed
    # Ensure setting bad count fails validation

    with pytest.raises(pystac.STACValidationError):
        PointcloudExtension.ext(pc_item).count = "not_an_int"  # type:ignore
        pc_item.validate()


@pytest.mark.vcr()
def test_type(pc_item: pystac.Item) -> None:
    # Get
    assert "pc:type" in pc_item.properties
    pc_type = PointcloudExtension.ext(pc_item).type
    assert pc_type == pc_item.properties["pc:type"]

    # Set
    PointcloudExtension.ext(pc_item).type = "sonar"
    assert "sonar" == pc_item.properties["pc:type"]

    # Validate
    pc_item.validate()


@pytest.mark.vcr()
def test_encoding(pc_item: pystac.Item) -> None:
    # Get
    assert "pc:encoding" in pc_item.properties
    pc_encoding = PointcloudExtension.ext(pc_item).encoding
    assert pc_encoding == pc_item.properties["pc:encoding"]

    # Set
    PointcloudExtension.ext(pc_item).encoding = "binary"
    assert "binary" == pc_item.properties["pc:encoding"]

    # Validate
    pc_item.validate()


@pytest.mark.vcr()
def test_schemas(pc_item: pystac.Item) -> None:
    # Get
    assert "pc:schemas" in pc_item.properties
    pc_schemas = [s.to_dict() for s in PointcloudExtension.ext(pc_item).schemas]
    assert pc_schemas == pc_item.properties["pc:schemas"]

    # Set
    schema = [Schema({"name": "X", "size": 8, "type": "floating"})]
    PointcloudExtension.ext(pc_item).schemas = schema
    assert [s.to_dict() for s in schema] == pc_item.properties["pc:schemas"]
    # Validate
    pc_item.validate()


@pytest.mark.vcr()
def test_statistics(pc_item: pystac.Item) -> None:
    # Get
    assert "pc:statistics" in pc_item.properties
    statistics = PointcloudExtension.ext(pc_item).statistics
    assert statistics is not None
    pc_statistics = [s.to_dict() for s in statistics]
    assert pc_statistics == pc_item.properties["pc:statistics"]

    # Set
    stats = [
        Statistic(
            {
                "average": 1,
                "count": 1,
                "maximum": 1,
                "minimum": 1,
                "name": "Test",
                "position": 1,
                "stddev": 1,
                "variance": 1,
            }
        )
    ]
    PointcloudExtension.ext(pc_item).statistics = stats
    assert [s.to_dict() for s in stats] == pc_item.properties["pc:statistics"]
    # Validate
    pc_item.validate()


@pytest.mark.vcr()
def test_density(pc_item: pystac.Item) -> None:
    # Get
    assert "pc:density" in pc_item.properties
    pc_density = PointcloudExtension.ext(pc_item).density
    assert pc_density == pc_item.properties["pc:density"]
    # Set
    density = 100
    PointcloudExtension.ext(pc_item).density = density
    assert density == pc_item.properties["pc:density"]
    # Validate
    pc_item.validate()


def test_pointcloud_schema() -> None:
    props: dict[str, Any] = {
        "name": "test",
        "size": 8,
        "type": "floating",
    }
    schema = Schema(props)
    assert props == schema.properties

    # test all getters and setters
    for k in props:
        if isinstance(props[k], str):
            val = props[k] + str(1)
        else:
            val = props[k] + 1
        setattr(schema, k, val)
        assert getattr(schema, k) == val

    schema = Schema.create("intensity", 16, SchemaType.UNSIGNED)
    assert schema.name == "intensity"
    assert schema.size == 16
    assert schema.type == "unsigned"

    with pytest.raises(STACError):
        schema.size = 0.5  # type: ignore

    empty_schema = Schema({})
    for required_prop in {"size", "name", "type"}:
        with pytest.raises(RequiredPropertyMissing):
            getattr(empty_schema, required_prop)


def test_pointcloud_statistics() -> None:
    props: dict[str, Any] = {
        "average": 1,
        "count": 1,
        "maximum": 1,
        "minimum": 1,
        "name": "Test",
        "position": 1,
        "stddev": 1,
        "variance": 1,
    }
    stat = Statistic(props)
    assert props == stat.properties

    # test all getters and setters
    for k in props:
        if isinstance(props[k], str):
            val = props[k] + str(1)
        else:
            val = props[k] + 1
        setattr(stat, k, val)
        assert getattr(stat, k) == val

    stat = Statistic.create("foo", 1, 2, 3, 4, 5, 6, 7)
    assert stat.name == "foo"
    assert stat.position == 1
    assert stat.average == 2
    assert stat.count == 3
    assert stat.maximum == 4
    assert stat.minimum == 5
    assert stat.stddev == 6
    assert stat.variance == 7

    stat.name = None  # type: ignore
    assert "name" not in stat.properties
    stat.position = None
    assert "position" not in stat.properties
    stat.average = None
    assert "average" not in stat.properties
    stat.count = None
    assert "count" not in stat.properties
    stat.maximum = None
    assert "maximum" not in stat.properties
    stat.minimum = None
    assert "minimum" not in stat.properties
    stat.stddev = None
    assert "stddev" not in stat.properties
    stat.variance = None
    assert "variance" not in stat.properties

    empty_stat = Statistic({})
    with pytest.raises(RequiredPropertyMissing):
        empty_stat.name


def test_statistics_accessor_when_no_stats(pc_no_stats_item: pystac.Item) -> None:
    assert PointcloudExtension.ext(pc_no_stats_item).statistics is None


def test_asset_extension(pc_no_stats_item: pystac.Item) -> None:
    asset = Asset(
        "https://github.com/PDAL/PDAL/blob"
        "/a6c986f68458e92414a66c664408bee4737bbb08/test/data/laz"
        "/autzen_trim.laz",
        "laz file",
        "The laz data",
        "application/octet-stream",
        ["data"],
        {"foo": "bar"},
    )
    pc_no_stats_item.add_asset("data", asset)
    ext = AssetPointcloudExtension(asset)
    assert ext.asset_href == asset.href
    assert ext.properties == asset.extra_fields
    assert ext.additional_read_properties == [pc_no_stats_item.properties]


def test_ext(pc_no_stats_item: pystac.Item) -> None:
    PointcloudExtension.ext(pc_no_stats_item)
    asset = Asset(
        "https://github.com/PDAL/PDAL/blob"
        "/a6c986f68458e92414a66c664408bee4737bbb08/test/data/laz"
        "/autzen_trim.laz",
        "laz file",
        "The laz data",
        "application/octet-stream",
        ["data"],
        {"foo": "bar"},
    )
    PointcloudExtension.ext(asset)

    class RandomObject:
        pass

    with pytest.raises(
        ExtensionTypeError,
        match=r"^PointcloudExtension does not apply to type 'RandomObject'$",
    ):
        # calling it wrong on purpose so -----------------v
        PointcloudExtension.ext(RandomObject())  # type: ignore


def test_extension_not_implemented(plain_item: pystac.Item) -> None:
    # Should raise exception if Item does not include extension URI
    with pytest.raises(pystac.ExtensionNotImplemented):
        _ = PointcloudExtension.ext(plain_item)

    # Should raise exception if owning Item does not include extension URI
    asset = plain_item.assets["thumbnail"]

    with pytest.raises(pystac.ExtensionNotImplemented):
        _ = PointcloudExtension.ext(asset)

    # Should succeed if Asset has no owner
    ownerless_asset = pystac.Asset.from_dict(asset.to_dict())
    _ = PointcloudExtension.ext(ownerless_asset)


def test_item_ext_add_to(plain_item: pystac.Item) -> None:
    assert PointcloudExtension.get_schema_uri() not in plain_item.stac_extensions

    _ = PointcloudExtension.ext(plain_item, add_if_missing=True)

    assert PointcloudExtension.get_schema_uri() in plain_item.stac_extensions


def test_asset_ext_add_to(plain_item: pystac.Item) -> None:
    assert PointcloudExtension.get_schema_uri() not in plain_item.stac_extensions
    asset = plain_item.assets["thumbnail"]

    _ = PointcloudExtension.ext(asset, add_if_missing=True)

    assert PointcloudExtension.get_schema_uri() in plain_item.stac_extensions


def test_summaries_count(collection: pystac.Collection) -> None:
    summaries_ext = PointcloudExtension.summaries(collection, True)
    count_range = RangeSummary(1000, 10000)

    summaries_ext.count = count_range

    assert summaries_ext.count == count_range

    summaries_dict = collection.to_dict()["summaries"]

    assert summaries_dict["pc:count"] == count_range.to_dict()


def test_summaries_type(collection: pystac.Collection) -> None:
    summaries_ext = PointcloudExtension.summaries(collection, True)
    type_list = [PhenomenologyType.LIDAR, "something"]

    summaries_ext.type = type_list

    assert summaries_ext.type == type_list

    summaries_dict = collection.to_dict()["summaries"]

    assert summaries_dict["pc:type"] == type_list


def test_summaries_encoding(collection: pystac.Collection) -> None:
    summaries_ext = PointcloudExtension.summaries(collection, True)
    encoding_list = ["LASzip"]

    summaries_ext.encoding = encoding_list

    assert summaries_ext.encoding == encoding_list

    summaries_dict = collection.to_dict()["summaries"]

    assert summaries_dict["pc:encoding"] == encoding_list


def test_summaries_density(collection: pystac.Collection) -> None:
    summaries_ext = PointcloudExtension.summaries(collection, True)
    density_range = RangeSummary(500.0, 1000.0)

    summaries_ext.density = density_range

    assert summaries_ext.density == density_range

    summaries_dict = collection.to_dict()["summaries"]

    assert summaries_dict["pc:density"] == density_range.to_dict()


def test_summaries_statistics(collection: pystac.Collection) -> None:
    summaries_ext = PointcloudExtension.summaries(collection, True)
    statistics_list = [
        Statistic(
            {
                "average": 637294.1783,
                "count": 10653336,
                "maximum": 639003.73,
                "minimum": 635577.79,
                "name": "X",
                "position": 0,
                "stddev": 967.9329805,
                "variance": 936894.2548,
            }
        )
    ]

    summaries_ext.statistics = statistics_list

    assert summaries_ext.statistics == statistics_list

    summaries_dict = collection.to_dict()["summaries"]

    assert summaries_dict["pc:statistics"] == [s.to_dict() for s in statistics_list]



from __future__ import annotations

import shutil
from copy import copy, deepcopy
from typing import TYPE_CHECKING, Any, Protocol, TypeVar

from pystac import MediaType, STACError, common_metadata, utils
from pystac.utils import is_absolute_href, make_absolute_href, make_relative_href

if TYPE_CHECKING:
    from pystac.common_metadata import CommonMetadata
    from pystac.extensions.ext import AssetExt

#: Generalized version of :class:`Asset`
A = TypeVar("A", bound="Asset")


class Asset:
    """An object that contains a link to data associated with an Item or Collection that
    can be downloaded or streamed.

    Args:
        href : Link to the asset object. Relative and absolute links are both
            allowed.
        title : Optional displayed title for clients and users.
        description : A description of the Asset providing additional details,
            such as how it was processed or created. CommonMark 0.29 syntax MAY be used
            for rich text representation.
        media_type : Optional description of the media type. Registered Media Types
            are preferred. See :class:`~pystac.MediaType` for common media types.
        roles : Optional, Semantic roles (i.e. thumbnail, overview,
            data, metadata) of the asset.
        extra_fields : Optional, additional fields for this asset. This is used
            by extensions as a way to serialize and deserialize properties on asset
            object JSON.
    """

    href: str
    """Link to the asset object. Relative and absolute links are both allowed."""

    title: str | None
    """Optional displayed title for clients and users."""

    description: str | None
    """A description of the Asset providing additional details, such as how it was
    processed or created. CommonMark 0.29 syntax MAY be used for rich text
    representation."""

    media_type: str | None
    """Optional description of the media type. Registered Media Types are preferred.
    See :class:`~pystac.MediaType` for common media types."""

    roles: list[str] | None
    """Optional, Semantic roles (i.e. thumbnail, overview, data, metadata) of the
    asset."""

    owner: Assets | None
    """The :class:`~pystac.Item` or :class:`~pystac.Collection` that this asset belongs
    to, or ``None`` if it has no owner."""

    extra_fields: dict[str, Any]
    """Optional, additional fields for this asset. This is used by extensions as a
    way to serialize and deserialize properties on asset object JSON."""

    def __init__(
        self,
        href: str,
        title: str | None = None,
        description: str | None = None,
        media_type: str | None = None,
        roles: list[str] | None = None,
        extra_fields: dict[str, Any] | None = None,
    ) -> None:
        self.href = utils.make_posix_style(href)
        self.title = title
        self.description = description
        self.media_type = media_type
        self.roles = roles
        self.extra_fields = extra_fields or {}

        # The Item which owns this Asset.
        self.owner = None

    def set_owner(self, obj: Assets) -> None:
        """Sets the owning item of this Asset.

        The owning item will be used to resolve relative HREFs of this asset.

        Args:
            obj: The Collection or Item that owns this asset.
        """
        self.owner = obj

    def get_absolute_href(self) -> str | None:
        """Gets the absolute href for this asset, if possible.

        If this Asset has no associated Item, and the asset HREF is a relative path,
            this method will return ``None``. If the Item that owns the Asset has no
            self HREF, this will also return ``None``.

        Returns:
            str: The absolute HREF of this asset, or None if an absolute HREF could not
                be determined.
        """
        item_self = self.owner.get_self_href() if self.owner is not None else None
        if utils.is_absolute_href(self.href, item_self):
            return self.href
        else:
            if item_self is not None:
                return utils.make_absolute_href(self.href, item_self)
            return None

    def to_dict(self) -> dict[str, Any]:
        """Returns this Asset as a dictionary.

        Returns:
            dict: A serialization of the Asset.
        """

        d: dict[str, Any] = {"href": self.href}

        if self.media_type is not None:
            d["type"] = self.media_type

        if self.title is not None:
            d["title"] = self.title

        if self.description is not None:
            d["description"] = self.description

        if self.extra_fields is not None and len(self.extra_fields) > 0:
            for k, v in self.extra_fields.items():
                d[k] = v

        if self.roles is not None:
            d["roles"] = self.roles

        return d

    def clone(self) -> Asset:
        """Clones this asset. Makes a ``deepcopy`` of the
        :attr:`~pystac.Asset.extra_fields`.

        Returns:
            Asset: The clone of this asset.
        """
        cls = self.__class__
        return cls(
            href=self.href,
            title=self.title,
            description=self.description,
            media_type=self.media_type,
            roles=self.roles,
            extra_fields=deepcopy(self.extra_fields),
        )

    def has_role(self, role: str) -> bool:
        """Check if a role exists in the Asset role list.

        Args:
            role: Role to check for existence.

        Returns:
            bool: True if role exists, else False.
        """
        if self.roles is None:
            return False
        else:
            return role in self.roles

    @property
    def common_metadata(self) -> CommonMetadata:
        """Access the asset's common metadata fields as a
        :class:`~pystac.CommonMetadata` object."""
        return common_metadata.CommonMetadata(self)

    def __repr__(self) -> str:
        return f"<Asset href={self.href}>"

    def _repr_html_(self) -> str:
        from html import escape

        from pystac.html.jinja_env import get_jinja_env

        jinja_env = get_jinja_env()
        if jinja_env:
            template = jinja_env.get_template("JSON.jinja2")
            return str(template.render(dict=self.to_dict(), plain=escape(repr(self))))
        else:
            return escape(repr(self))

    @classmethod
    def from_dict(cls: type[A], d: dict[str, Any]) -> A:
        """Constructs an Asset from a dict.

        Returns:
            Asset: The Asset deserialized from the JSON dict.
        """
        d = copy(d)
        href = d.pop("href")
        media_type = d.pop("type", None)
        title = d.pop("title", None)
        description = d.pop("description", None)
        roles = d.pop("roles", None)
        properties = None
        if any(d):
            properties = d

        return cls(
            href=href,
            media_type=media_type,
            title=title,
            description=description,
            roles=roles,
            extra_fields=properties,
        )

    def move(self, href: str) -> Asset:
        """Moves this asset's file to a new location on the local filesystem,
        setting the asset href accordingly.

        Modifies the asset in place, and returns the same asset.

        Args:
            href: The new asset location. Must be a local path. If relative
                it must be relative to the owner object.

        Returns:
            Asset: The asset with the updated href.
        """
        src = _absolute_href(self.href, self.owner, "move")
        dst = _absolute_href(href, self.owner, "move")
        shutil.move(src, dst)
        self.href = href
        return self

    def copy(self, href: str) -> Asset:
        """Copies this asset's file to a new location on the local filesystem,
        setting the asset href accordingly.

        Modifies the asset in place, and returns the same asset.

        Args:
            href: The new asset location. Must be a local path. If relative
                it must be relative to the owner object.

        Returns:
            Asset: The asset with the updated href.
        """
        src = _absolute_href(self.href, self.owner, "copy")
        dst = _absolute_href(href, self.owner, "copy")
        shutil.copy2(src, dst)
        self.href = href
        return self

    def delete(self) -> None:
        """Delete this asset's file. Does not delete the asset from the item
        that owns it. See :meth:`~pystac.Item.delete_asset` for that.

        Does not modify the asset.
        """
        import os

        href = _absolute_href(self.href, self.owner, "delete")
        os.remove(href)

    @property
    def ext(self) -> AssetExt:
        """Accessor for extension classes on this asset

        Example::

            asset.ext.proj.code = "EPSG:4326"
        """
        from pystac.extensions.ext import AssetExt

        return AssetExt(stac_object=self)


class Assets(Protocol):
    """Protocol, with functionality, for STAC objects that have assets."""

    assets: dict[str, Asset]
    """The asset dictionary."""

    def get_assets(
        self,
        media_type: str | MediaType | None = None,
        role: str | None = None,
    ) -> dict[str, Asset]:
        """Get this object's assets.

        Args:
            media_type: If set, filter the assets such that only those with a
                matching ``media_type`` are returned.
            role: If set, filter the assets such that only those with a matching
                ``role`` are returned.

        Returns:
            Dict[str, Asset]: A dictionary of assets that match ``media_type``
                and/or ``role`` if set or else all of this object's assets.
        """
        return {
            k: deepcopy(v)
            for k, v in self.assets.items()
            if (media_type is None or v.media_type == media_type)
            and (role is None or v.has_role(role))
        }

    def add_asset(self, key: str, asset: Asset) -> None:
        """Adds an Asset to this object.

        Args:
            key : The unique key of this asset.
            asset : The Asset to add.
        """
        asset.set_owner(self)
        self.assets[key] = asset

    def delete_asset(self, key: str) -> None:
        """Deletes the asset at the given key, and removes the asset's data
        file from the local filesystem.

        It is an error to attempt to delete an asset's file if it is on a
        remote filesystem.

        To delete the asset without removing the file, use `del item.assets["key"]`.

        Args:
            key: The unique key of this asset.
        """
        asset = self.assets[key]
        asset.set_owner(self)
        asset.delete()

        del self.assets[key]

    def make_asset_hrefs_relative(self) -> Assets:
        """Modify each asset's HREF to be relative to this object's self HREF.

        Returns:
            Item: self
        """
        self_href = self.get_self_href()
        for asset in self.assets.values():
            if is_absolute_href(asset.href, self_href):
                if self_href is None:
                    raise STACError(
                        "Cannot make asset HREFs relative if no self_href is set."
                    )
                asset.href = make_relative_href(asset.href, self_href)
        return self

    def make_asset_hrefs_absolute(self) -> Assets:
        """Modify each asset's HREF to be absolute.

        Any asset HREFs that are relative will be modified to absolute based on this
        item's self HREF.

        Returns:
            Assets: self
        """
        self_href = self.get_self_href()
        for asset in self.assets.values():
            if not is_absolute_href(asset.href, self_href):
                if self_href is None:
                    raise STACError(
                        "Cannot make relative asset HREFs absolute "
                        "if no self_href is set."
                    )
                asset.href = make_absolute_href(asset.href, self_href)
        return self

    def get_self_href(self) -> str | None:
        """Abstract definition of STACObject.get_self_href.

        Needed to make the `make_asset_hrefs_{absolute|relative}` methods pass
        type checking. Refactoring out all the link behavior in STACObject to
        its own protocol would be too heavy, so we just use this stub instead.
        """
        ...


def _absolute_href(href: str, owner: Assets | None, action: str = "access") -> str:
    item_self = owner.get_self_href() if owner else None
    if utils.is_absolute_href(href, item_self):
        return href
    else:
        if item_self is None:
            raise ValueError(
                f"Cannot {action} file if asset href ('{href}') is relative "
                "and owner item is not set. Hint: try using "
                ":func:`~pystac.Item.make_asset_hrefs_absolute`"
            )
        return utils.make_absolute_href(href, item_self)



import tempfile
from typing import Any

import pytest

import pystac
from pystac import HIERARCHICAL_LINKS, Catalog, CatalogType, Collection
from pystac.utils import is_absolute_href, make_absolute_href
from pystac.validation import validate_dict
from tests.utils import TestCases

CTYPES = [
    CatalogType.ABSOLUTE_PUBLISHED,
    CatalogType.RELATIVE_PUBLISHED,
    CatalogType.SELF_CONTAINED,
]


class TestSTACWriting:
    """Tests writing STACs, using JSON Schema validation,
    and ensure that links are correctly set to relative or absolute.
    """

    def validate_catalog(self, catalog: Catalog) -> int:
        catalog.validate()
        validated_count = 1

        for child in catalog.get_children():
            validated_count += self.validate_catalog(child)

        for item in catalog.get_items():
            item.validate()
            validated_count += 1

        return validated_count

    def validate_file(self, path: str, object_type: str) -> list[Any]:
        d = pystac.StacIO.default().read_json(path)
        return validate_dict(d, pystac.STACObjectType(object_type))

    def validate_link_types(self, root_href: str, catalog_type: CatalogType) -> None:
        def validate_item_link_type(
            href: str, link_type: str, should_include_self: bool
        ) -> None:
            item_dict = pystac.StacIO.default().read_json(href)
            item = pystac.Item.from_file(href)
            rel_links = [
                *HIERARCHICAL_LINKS,
                *pystac.EXTENSION_HOOKS.get_extended_object_links(item),
            ]
            for link in item.get_links():
                if not link.rel == "self":
                    if link_type == "RELATIVE" and link.rel in rel_links:
                        assert not is_absolute_href(link.href)
                    else:
                        assert is_absolute_href(link.href)

            rels = {link["rel"] for link in item_dict["links"]}
            assert ("self" in rels) == should_include_self

        def validate_catalog_link_type(
            href: str, link_type: str, should_include_self: bool
        ) -> None:
            cat_dict = pystac.StacIO.default().read_json(href)
            cat = pystac.read_file(href)
            assert isinstance(cat, Catalog)

            rels = {link["rel"] for link in cat_dict["links"]}
            assert ("self" in rels) == should_include_self

            for child_link in cat.get_child_links():
                child_href = make_absolute_href(child_link.href, href)
                validate_catalog_link_type(
                    child_href,
                    link_type,
                    catalog_type == CatalogType.ABSOLUTE_PUBLISHED,
                )

            for item_link in cat.get_item_links():
                item_href = make_absolute_href(item_link.href, href)
                validate_item_link_type(
                    item_href, link_type, catalog_type == CatalogType.ABSOLUTE_PUBLISHED
                )

        link_type = "RELATIVE"
        if catalog_type == CatalogType.ABSOLUTE_PUBLISHED:
            link_type = "ABSOLUTE"

        root_should_include_href = catalog_type in [
            CatalogType.ABSOLUTE_PUBLISHED,
            CatalogType.RELATIVE_PUBLISHED,
        ]

        validate_catalog_link_type(root_href, link_type, root_should_include_href)

    def do_test(self, catalog: Catalog, catalog_type: CatalogType) -> None:
        with tempfile.TemporaryDirectory() as tmp_dir:
            catalog.normalize_hrefs(tmp_dir)
            self.validate_catalog(catalog)

            catalog.save(catalog_type=catalog_type)

            root_href = catalog.self_href
            self.validate_link_types(root_href, catalog_type)

            for parent, _, items in catalog.walk():
                if issubclass(type(parent), Collection):
                    stac_object_type = pystac.STACObjectType.COLLECTION
                else:
                    stac_object_type = pystac.STACObjectType.CATALOG
                self.validate_file(parent.self_href, stac_object_type)

                for item in items:
                    self.validate_file(item.self_href, pystac.STACObjectType.ITEM)

    @pytest.mark.vcr()
    @pytest.mark.parametrize("catalog", TestCases.all_test_catalogs())
    @pytest.mark.parametrize("catalog_type", CTYPES)
    def test_testcases(self, catalog: Catalog, catalog_type: CatalogType) -> None:
        catalog = catalog.full_copy()
        self.do_test(catalog, catalog_type)
